import unittest
from unittest import TestCase, mock
from unittest.mock import Mock, MagicMock, patch, call
import json

from data_vault import get_synced_params, VaultEnvironment
from helper import KeeperApiHelper
from keepercommander.commands.device_management import (
    DeviceUserListCommand,
    DeviceUserActionCommand,
    DeviceUserRenameCommand,
    DeviceAdminListCommand,
    DeviceAdminActionCommand,
    ValidationError
)
from keepercommander.proto import DeviceManagement_pb2, APIRequest_pb2
from keepercommander.error import KeeperApiError


class MockDevice:
    """Mock device object for testing."""
    
    def __init__(self, device_name="Test Device", device_id=1, 
                 client_type=DeviceManagement_pb2.CHROME_EXTENSION,
                 client_type_category=DeviceManagement_pb2.CAT_EXTENSION,
                 client_form_factor=APIRequest_pb2.FF_EMPTY,
                 login_state=APIRequest_pb2.LOGGED_IN,
                 device_status=APIRequest_pb2.DEVICE_OK,
                 last_modified_time=1640995200000,
                 encrypted_device_token=None):
        self.deviceName = device_name
        self.clientType = client_type
        self.clientTypeCategory = client_type_category
        self.clientFormFactor = client_form_factor
        self.loginState = login_state
        self.deviceStatus = device_status
        self.lastModifiedTime = last_modified_time
        self.encryptedDeviceToken = encrypted_device_token or f'token_{device_id}'.encode()


class MockDeviceGroup:
    """Mock device group for testing."""
    
    def __init__(self, devices):
        self.devices = devices


class MockDeviceUserGroup:
    """Mock device user group for testing."""
    
    def __init__(self, enterprise_user_id, device_groups):
        self.enterpriseUserId = enterprise_user_id
        self.deviceGroups = device_groups


class MockDeviceResponse:
    """Mock device response for testing."""
    
    def __init__(self, device_groups):
        self.deviceGroups = device_groups


class MockAdminDeviceResponse:
    """Mock admin device response for testing."""
    
    def __init__(self, device_user_list):
        self.deviceUserList = device_user_list


class TestDeviceUserListCommand(TestCase):
    """Test cases for DeviceUserListCommand."""

    def setUp(self):
        """Set up test fixtures."""
        self.command = DeviceUserListCommand()
        self.params = get_synced_params()
        
        self.device1 = MockDevice("Test Device 1", 1, encrypted_device_token=b'token1')
        self.device2 = MockDevice("Test Device 2", 2, encrypted_device_token=b'token2')
        
        self.communicate_mock = mock.patch('keepercommander.api.communicate_rest').start()

    def tearDown(self):
        """Clean up after tests."""
        mock.patch.stopall()

    def test_execute_success_table_format(self):
        """Test successful execution with table format."""
        device_group = MockDeviceGroup([self.device1, self.device2])
        mock_response = MockDeviceResponse([device_group])
        self.communicate_mock.return_value = mock_response
        
        with patch.object(self.command, '_display_table') as mock_display:
            self.command.execute(self.params, format='table')
            
            self.communicate_mock.assert_called_once_with(
                self.params, None, 'dm/device_user_list', rs_type=DeviceManagement_pb2.DeviceUserResponse
            )
            
            mock_display.assert_called_once()
            devices = mock_display.call_args[0][0]
            self.assertEqual(len(devices), 2)

    def test_execute_success_json_format(self):
        """Test successful execution with JSON format."""
        device_group = MockDeviceGroup([self.device1])
        mock_response = MockDeviceResponse([device_group])
        self.communicate_mock.return_value = mock_response
        
        with patch.object(self.command, '_display_json') as mock_display:
            self.command.execute(self.params, format='json')
            
            mock_display.assert_called_once()

    def test_execute_no_devices(self):
        """Test execution when no devices are found."""
        mock_response = MockDeviceResponse([])
        self.communicate_mock.return_value = mock_response
        
        with patch('builtins.print') as mock_print:
            self.command.execute(self.params)
            
            mock_print.assert_called_with("No devices found.")

    def test_execute_api_error_handled_gracefully(self):
        """Test execution when API error is handled gracefully."""
        self.communicate_mock.return_value = None  
        
        result = self.command.execute(self.params)
        self.assertIsNone(result)

    def test_display_table_format(self):
        """Test table display format."""
        devices = [self.device1, self.device2]
        
        with patch('builtins.print'):  
            self.command._display_table(devices)
        
        

    def test_display_json_format(self):
        """Test JSON display format."""
        devices = [self.device1, self.device2]
        
        with patch('builtins.print') as mock_print:
            self.command._display_json(devices)
            
            mock_print.assert_called_once()
            
            printed_json = mock_print.call_args[0][0]
            data = json.loads(printed_json)
            
            self.assertIn('devices', data)
            self.assertEqual(len(data['devices']), 2)
            self.assertEqual(data['devices'][0]['deviceName'], "Test Device 1")


class TestDeviceUserActionCommand(TestCase):
    """Test cases for DeviceUserActionCommand."""

    def setUp(self):
        """Set up test fixtures."""
        self.command = DeviceUserActionCommand()
        self.params = get_synced_params()
        
        self.device1 = MockDevice("Test Device 1", 1, encrypted_device_token=b'token1')
        self.device2 = MockDevice("Test Device 2", 2, encrypted_device_token=b'token2')
        
        self.communicate_mock = mock.patch('keepercommander.api.communicate_rest').start()

    def tearDown(self):
        """Clean up after tests."""
        mock.patch.stopall()

    def test_validate_inputs_success(self):
        """Test successful input validation."""
        kwargs = {
            'action': 'logout',
            'devices': ['1', '2']
        }
        
        self.command._validate_inputs(**kwargs)
        
        self.assertEqual(kwargs['devices'], ['1', '2'])

    def test_validate_inputs_invalid_action(self):
        """Test validation with invalid action."""
        kwargs = {
            'action': 'invalid_action',
            'devices': ['1']
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("Invalid action", str(cm.exception))

    def test_validate_inputs_no_devices(self):
        """Test validation with no devices."""
        kwargs = {
            'action': 'logout',
            'devices': []
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("At least one device must be specified", str(cm.exception))

    def test_validate_inputs_link_minimum_devices(self):
        """Test validation for link action requiring minimum 2 devices."""
        kwargs = {
            'action': 'link',
            'devices': ['1']  
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("requires at least 2 devices", str(cm.exception))

    def test_validate_inputs_unlink_minimum_devices(self):
        """Test validation for unlink action requiring minimum 2 devices."""
        kwargs = {
            'action': 'unlink',
            'devices': ['1']  
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("requires at least 2 devices", str(cm.exception))

    def test_execute_success(self):
        """Test successful execution of device action."""
        device_group = MockDeviceGroup([self.device1, self.device2])
        list_response = MockDeviceResponse([device_group])
        
        action_result = Mock()
        action_result.deviceActionStatus = DeviceManagement_pb2.SUCCESS
        action_result.encryptedDeviceToken = [b'token1']
        
        action_response = Mock()
        action_response.deviceActionResult = [action_result]
        
        updated_list_response = MockDeviceResponse([device_group])
        
        self.communicate_mock.side_effect = [list_response, action_response, updated_list_response]
        
        kwargs = {
            'action': 'logout',
            'devices': ['1']
        }
        
        with patch.object(self.command, '_display_results') as mock_display, \
             patch.object(self.command, '_show_updated_device_list') as mock_show_list:
            
            self.command.execute(self.params, **kwargs)
            
            self.assertGreaterEqual(self.communicate_mock.call_count, 2)
            
            mock_display.assert_called_once()

    def test_execute_validation_error(self):
        """Test execution with validation error."""
        kwargs = {
            'action': 'invalid',
            'devices': ['1']
        }
        
        with patch('logging.error') as mock_log:
            result = self.command.execute(self.params, **kwargs)
            
            self.assertIsNone(result)

    def test_execute_no_matching_devices(self):
        """Test execution when no devices match identifiers."""
        device_group = MockDeviceGroup([self.device1])
        list_response = MockDeviceResponse([device_group])
        self.communicate_mock.return_value = list_response
        
        kwargs = {
            'action': 'logout',
            'devices': ['999']  
        }
        
        with patch('builtins.print') as mock_print:
            self.command.execute(self.params, **kwargs)
            
            mock_print.assert_called_with("No matching devices found.")

    def test_get_action_parser(self):
        """Test getting action-specific parser."""
        logout_parser = self.command.get_action_parser('logout')
        self.assertIsNotNone(logout_parser)
        self.assertEqual(logout_parser.prog, 'device-action logout')
        
        invalid_parser = self.command.get_action_parser('invalid_action')
        self.assertIsNone(invalid_parser)

    def test_get_action_verb(self):
        """Test action verb mapping."""
        test_cases = [
            ('logout', 'logged out'),
            ('remove', 'removed'),
            ('lock', 'locked'),
            ('unlock', 'unlocked'),
            ('account-lock', 'account locked'),
            ('account-unlock', 'account unlocked'),
            ('link', 'linked'),
            ('unlink', 'unlinked'),
            ('unknown_action', 'unknown_actioned')
        ]
        
        for action, expected_verb in test_cases:
            with self.subTest(action=action):
                result = self.command._get_action_verb(action)
                self.assertEqual(result, expected_verb)


class TestDeviceUserRenameCommand(TestCase):
    """Test cases for DeviceUserRenameCommand."""

    def setUp(self):
        """Set up test fixtures."""
        self.command = DeviceUserRenameCommand()
        self.params = get_synced_params()
        
        self.device1 = MockDevice("Old Device Name", 1, encrypted_device_token=b'token1')
        
        self.communicate_mock = mock.patch('keepercommander.api.communicate_rest').start()

    def tearDown(self):
        """Clean up after tests."""
        mock.patch.stopall()

    def test_validate_inputs_success(self):
        """Test successful input validation."""
        kwargs = {
            'device': 'device1',
            'new_name': 'New Device Name'
        }
        
        self.command._validate_inputs(**kwargs)
        
        self.assertEqual(kwargs['new_name'], 'New Device Name')

    def test_validate_inputs_sanitize_name(self):
        """Test input validation with name sanitization."""
        kwargs = {
            'device': 'device1',
            'new_name': 'Device<script>alert("xss")</script>Name'
        }
        
       
        try:
            self.command._validate_inputs(**kwargs)
            self.assertTrue(True)
        except ValidationError:
            self.fail("Validation should not fail for sanitizable input")

    def test_validate_inputs_invalid_device(self):
        """Test validation with invalid device identifier."""
        kwargs = {
            'device': '',  
            'new_name': 'New Name'
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("Device identifier is required", str(cm.exception))

    def test_validate_inputs_empty_name(self):
        """Test validation with empty name."""
        kwargs = {
            'device': 'device1',
            'new_name': ''
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("New device name is required", str(cm.exception))

    def test_validate_inputs_name_only_invalid_chars(self):
        """Test validation with name containing only invalid characters."""
        kwargs = {
            'device': 'device1',
            'new_name': '<>"\'<>'  
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("contains only invalid characters", str(cm.exception))

    def test_execute_success(self):
        """Test successful device rename."""
        device_group = MockDeviceGroup([self.device1])
        list_response = MockDeviceResponse([device_group])
        
        rename_result = Mock()
        rename_result.deviceActionStatus = DeviceManagement_pb2.SUCCESS
        rename_result.encryptedDeviceToken = [b'token1']
        
        rename_response = Mock()
        rename_response.deviceRenameResult = [rename_result]
        
        updated_list_response = MockDeviceResponse([device_group])
        
        self.communicate_mock.side_effect = [list_response, rename_response, updated_list_response]
        
        kwargs = {
            'device': '1',
            'new_name': 'New Device Name'
        }
        
        with patch.object(self.command, '_display_results') as mock_display, \
             patch.object(self.command, '_show_updated_device_list') as mock_show_list:
            
            self.command.execute(self.params, **kwargs)
            
            self.assertGreaterEqual(self.communicate_mock.call_count, 2)
            
            mock_display.assert_called_once()

    def test_execute_validation_error(self):
        """Test execution with validation error."""
        kwargs = {
            'device': '',  
            'new_name': 'New Name'
        }
        
        with patch('logging.error') as mock_log:
            result = self.command.execute(self.params, **kwargs)
            
            self.assertIsNone(result)
            mock_log.assert_called()

    def test_execute_no_matching_device(self):
        """Test execution when device is not found."""
        device_group = MockDeviceGroup([self.device1])
        list_response = MockDeviceResponse([device_group])
        self.communicate_mock.return_value = list_response
        
        kwargs = {
            'device': '999',  
            'new_name': 'New Name'
        }
        
        with patch('builtins.print') as mock_print:
            self.command.execute(self.params, **kwargs)
            
            mock_print.assert_called_with("No matching device found.")


class TestDeviceAdminListCommand(TestCase):
    """Test cases for DeviceAdminListCommand."""

    def setUp(self):
        """Set up test fixtures."""
        self.command = DeviceAdminListCommand()
        self.params = get_synced_params()
        
        # Mock devices
        self.device1 = MockDevice("User 1 Device", 1, encrypted_device_token=b'token1')
        self.device2 = MockDevice("User 2 Device", 2, encrypted_device_token=b'token2')
        
        # Mock API communication
        self.communicate_mock = mock.patch('keepercommander.api.communicate_rest').start()

    def tearDown(self):
        """Clean up after tests."""
        mock.patch.stopall()

    def test_validate_inputs_success(self):
        """Test successful input validation."""
        kwargs = {
            'enterprise_user_ids': [123, 456]
        }
        
        self.command._validate_inputs(**kwargs)

    def test_validate_inputs_no_user_ids(self):
        """Test validation with no user IDs."""
        kwargs = {
            'enterprise_user_ids': []
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("Enterprise User ID is required", str(cm.exception))
        self.assertIn("ei --users", str(cm.exception))

    def test_validate_inputs_invalid_user_id(self):
        """Test validation with invalid user ID."""
        kwargs = {
            'enterprise_user_ids': [123, -1]  # -1 is invalid
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("Invalid enterprise user ID: -1", str(cm.exception))


    def test_execute_success(self):
        """Test successful execution of admin device list."""
        device_group1 = MockDeviceGroup([self.device1])
        device_user_group1 = MockDeviceUserGroup(123, [device_group1])
        
        device_group2 = MockDeviceGroup([self.device2])
        device_user_group2 = MockDeviceUserGroup(456, [device_group2])
        
        mock_response = MockAdminDeviceResponse([device_user_group1, device_user_group2])
        self.communicate_mock.return_value = mock_response
        
        kwargs = {
            'enterprise_user_ids': [123, 456]
        }
        
        with patch.object(self.command, '_display_results') as mock_display:
            self.command.execute(self.params, **kwargs)
            
            self.communicate_mock.assert_called_once()
            
            mock_display.assert_called_once()
            devices = mock_display.call_args[0][0]
            self.assertEqual(len(devices), 2)  

    def test_execute_validation_error(self):
        """Test execution with validation error."""
        kwargs = {
            'enterprise_user_ids': [] 
        }
        
        with patch('logging.error') as mock_log:
            result = self.command.execute(self.params, **kwargs)
            
            self.assertIsNone(result)
            mock_log.assert_called()

    def test_execute_no_devices(self):
        """Test execution when no devices are found."""
        mock_response = MockAdminDeviceResponse([])
        self.communicate_mock.return_value = mock_response
        
        kwargs = {
            'enterprise_user_ids': [123]
        }
        
        with patch('builtins.print') as mock_print:
            self.command.execute(self.params, **kwargs)
            
            mock_print.assert_called_with("No devices found.")


class TestDeviceAdminActionCommand(TestCase):
    """Test cases for DeviceAdminActionCommand."""

    def setUp(self):
        """Set up test fixtures."""
        self.command = DeviceAdminActionCommand()
        self.params = get_synced_params()
        
        self.device1 = MockDevice("User Device", 1, encrypted_device_token=b'token1')
        
        self.communicate_mock = mock.patch('keepercommander.api.communicate_rest').start()

    def tearDown(self):
        """Clean up after tests."""
        mock.patch.stopall()

    def test_validate_inputs_success(self):
        """Test successful input validation."""
        kwargs = {
            'action': 'logout',
            'enterprise_user_id': 123,
            'devices': ['1', '2']
        }
        
        self.command._validate_inputs(**kwargs)
        
        self.assertEqual(kwargs['devices'], ['1', '2'])

    def test_validate_inputs_invalid_action(self):
        """Test validation with invalid action."""
        kwargs = {
            'action': 'invalid_action',
            'enterprise_user_id': 123,
            'devices': ['1']
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("Invalid action", str(cm.exception))

    def test_validate_inputs_invalid_user_id(self):
        """Test validation with invalid enterprise user ID."""
        kwargs = {
            'action': 'logout',
            'enterprise_user_id': -1,  
            'devices': ['1']
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("Invalid enterprise user ID", str(cm.exception))

    def test_validate_inputs_no_devices(self):
        """Test validation with no devices."""
        kwargs = {
            'action': 'logout',
            'enterprise_user_id': 123,
            'devices': []
        }
        
        with self.assertRaises(ValidationError) as cm:
            self.command._validate_inputs(**kwargs)
        
        self.assertIn("At least one device must be specified", str(cm.exception))

    def test_execute_success(self):
        """Test successful execution of admin device action."""
        device_group = MockDeviceGroup([self.device1])
        device_user_group = MockDeviceUserGroup(123, [device_group])
        list_response = MockAdminDeviceResponse([device_user_group])
        
        action_result = Mock()
        action_result.deviceActionStatus = DeviceManagement_pb2.SUCCESS
        action_result.encryptedDeviceToken = [b'token1']
        
        action_response = Mock()
        action_response.deviceAdminActionResults = [action_result]
        
        updated_list_response = MockAdminDeviceResponse([device_user_group])
        
        self.communicate_mock.side_effect = [list_response, action_response, updated_list_response]
        
        kwargs = {
            'action': 'logout',
            'enterprise_user_id': 123,
            'devices': ['1']
        }
        
        with patch.object(self.command, '_display_results') as mock_display, \
             patch.object(self.command, '_show_updated_device_list') as mock_show_list:
            
            self.command.execute(self.params, **kwargs)
            
            self.assertGreaterEqual(self.communicate_mock.call_count, 2)
            
            mock_display.assert_called_once()

    def test_execute_validation_error(self):
        """Test execution with validation error."""
        kwargs = {
            'action': 'invalid',
            'enterprise_user_id': 123,
            'devices': ['1']
        }
        
        with patch('logging.error') as mock_log:
            result = self.command.execute(self.params, **kwargs)
            
            self.assertIsNone(result)

    def test_execute_no_matching_devices(self):
        """Test execution when no devices match identifiers."""
        device_group = MockDeviceGroup([self.device1])
        device_user_group = MockDeviceUserGroup(123, [device_group])
        list_response = MockAdminDeviceResponse([device_user_group])
        self.communicate_mock.return_value = list_response
        
        kwargs = {
            'action': 'logout',
            'enterprise_user_id': 123,
            'devices': ['999']  
        }
        
        with patch('builtins.print') as mock_print:
            self.command.execute(self.params, **kwargs)
            
            mock_print.assert_called_with("No matching devices found.")

    def test_get_action_parser(self):
        """Test getting action-specific parser."""
        logout_parser = self.command.get_action_parser('logout')
        self.assertIsNotNone(logout_parser)
        self.assertEqual(logout_parser.prog, 'device-admin-action logout')
        
        # Test non-existent action
        invalid_parser = self.command.get_action_parser('invalid_action')
        self.assertIsNone(invalid_parser)

    def test_get_action_verb(self):
        """Test action verb mapping for admin actions."""
        test_cases = [
            ('logout', 'logged out'),
            ('remove', 'removed'),
            ('lock', 'locked'),
            ('unlock', 'unlocked'),
            ('account-lock', 'account locked'),
            ('account-unlock', 'account unlocked'),
            ('unknown_action', 'unknown_actioned')
        ]
        
        for action, expected_verb in test_cases:
            with self.subTest(action=action):
                result = self.command._get_action_verb(action)
                self.assertEqual(result, expected_verb)






import re

""" Filter rows by pattern(s). All non-primitive data type values contained are assumed to be of type List """
def filter_rows(rows, patterns, use_regex=False):
    if not patterns:
        return rows

    is_a = lambda o, t: isinstance(o, t)
    to_string = lambda el: el if is_a(el, str) else is_a(el, list) and ' '.join([to_string(c) for c in el]) or str(el) or ''
    if use_regex:
        pattern = re.compile('|'.join(patterns))
        is_match = lambda row: bool(pattern.search(row))
    else:
        is_match = lambda row: any(p for p in patterns if p.casefold() in row.casefold())
    return list(
        filter(
            lambda row: is_match(to_string(row)),
            rows
        )
    )



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from typing import Dict, Any, Optional
from ..decorators.logging import logger, debug_decorator
from .service_config import ServiceConfig
from ..util.tunneling import generate_ngrok_url

class NgrokConfigurator:
    @staticmethod
    @debug_decorator
    def configure_ngrok(config_data: Dict[str, Any], service_config: ServiceConfig) -> Optional[int]:
        """Configure ngrok if enabled. Returns ngrok PID if started in background mode, None otherwise."""
        if config_data.get("ngrok") == 'y':
            logger.debug("Configuring ngrok tunnel")
            result = generate_ngrok_url(
                config_data["port"], 
                config_data["ngrok_auth_token"],
                config_data["ngrok_custom_domain"],
                config_data["run_mode"],
            )
            if isinstance(result, tuple):
                config_data["ngrok_public_url"], ngrok_pid = result
                if config_data["ngrok_public_url"]:
                    print(f'Generated ngrok URL: {config_data["ngrok_public_url"]}')
                else:
                    print('Ngrok tunnel started, URL will be available in logs')
                return ngrok_pid
            else:
                config_data["ngrok_public_url"] = result
                print(f'Generated ngrok URL: {result}')
                return None
            # service_config.save_config(config_data)
        return None


from unittest import TestCase, mock
import logging

from data_vault import get_synced_params
from helper import KeeperApiHelper
from keepercommander.commands import folder
from keepercommander.error import CommandError


class TestFolder(TestCase):
    def setUp(self):
        self.communicate_mock = mock.patch('keepercommander.api.communicate').start()
        self.communicate_mock.side_effect = KeeperApiHelper.communicate_command

    def tearDown(self):
        mock.patch.stopall()

    def test_list(self):
        params = get_synced_params()

        cmd = folder.FolderListCommand()
        with mock.patch('builtins.print'), mock.patch('keepercommander.api.get_record_shares'):
            cmd.execute(params)
            cmd.execute(params, detail=True)
            
    def test_list_json_no_ansi_colors(self):
        """Test that ls command with JSON format doesn't include ANSI color codes."""
        params = get_synced_params()
        cmd = folder.FolderListCommand()
        
        # Execute with JSON format - this should not include ANSI color codes
        with mock.patch('keepercommander.api.get_record_shares'):
            result = cmd.execute(params, detail=True, format='json')
            
            # If result is returned (JSON/CSV format), check it doesn't contain ANSI codes
            if result:
                import json
                # Convert to JSON string to check for ANSI escape sequences
                json_str = json.dumps(result)
                # Check that no ANSI escape sequences are present
                self.assertNotIn('\u001b[', json_str, "JSON output should not contain ANSI escape sequences")
                self.assertNotIn('\\u001b[', json_str, "JSON output should not contain ANSI escape sequences")

    def test_change_directory(self):
        params = get_synced_params()

        cmd = folder.FolderCdCommand()
        folder_name = next(iter([x['name_unencrypted'] for x in params.shared_folder_cache.values()]))
        owd = params.current_folder
        cmd.execute(params, folder=folder_name)
        self.assertNotEqual(owd, params.current_folder)

    def test_tree(self):
        params = get_synced_params()
        cmd = folder.FolderTreeCommand()

        with mock.patch('builtins.print'):
            cmd.execute(params)
        with self.assertRaises(CommandError):
            cmd.execute(params, folder='Invalid')

    def test_make_folder(self):
        params = get_synced_params()
        cmd = folder.FolderMakeCommand()

        def is_user_folder(rq):
            self.assertEqual(rq['command'], 'folder_add')
            self.assertEqual(rq['folder_type'], 'user_folder')

        def is_shared_folder(rq):
            self.assertEqual(rq['command'], 'folder_add')
            self.assertEqual(rq['folder_type'], 'shared_folder')

        def is_shared_folder_folder(rq):
            self.assertEqual(rq['command'], 'folder_add')
            self.assertEqual(rq['folder_type'], 'shared_folder_folder')

        KeeperApiHelper.communicate_expect([is_user_folder])
        cmd.execute(params, user_folder=True, folder='New Folder')

        KeeperApiHelper.communicate_expect([is_shared_folder])
        cmd.execute(params, shared_folder=True, folder='New Shared Folder')

        with mock.patch('keepercommander.commands.folder.user_choice') as mock_choice:
            mock_choice.return_value = 'n'
            KeeperApiHelper.communicate_expect([is_user_folder])
            cmd.execute(params, folder='New Folder')
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            mock_choice.return_value = 'y'
            KeeperApiHelper.communicate_expect([is_shared_folder])
            cmd.execute(params, folder='New Shared Folder')
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            with mock.patch('builtins.input') as mock_input:
                mock_choice.return_value = 'n'
                mock_input.return_value = 'New Personal Folder'
                KeeperApiHelper.communicate_expect([is_user_folder])
                cmd.execute(params)
                self.assertTrue(KeeperApiHelper.is_expect_empty())

        shared_folder = next(iter([x for x in params.folder_cache.values() if x.type == 'shared_folder']))
        with self.assertLogs(level=logging.WARNING):
            cmd.execute(params, folder=shared_folder.name)

        params.current_folder = shared_folder.uid
        KeeperApiHelper.communicate_expect([is_shared_folder_folder])
        cmd.execute(params, folder='New SubFolder')
        self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_remove_non_existing_folder(self):
        params = get_synced_params()
        cmd = folder.FolderRemoveCommand()

        with self.assertRaises(CommandError):
            cmd.execute(params, folder='Invalid Name')

    def test_delete_folders(self):
        params = get_synced_params()
        cmd = folder.FolderRemoveCommand()
        all_folders = [x for x in params.folder_cache.values() if not x.parent_uid]

        def pre_delete(rq):
            self.assertEqual(rq['command'], 'pre_delete')
            return {
                'pre_delete_response': {
                    'would_delete': {
                        'deletion_summary': ['All root folders']
                    },
                    'pre_delete_token': 'token'
                }
            }

        with mock.patch('builtins.print'), mock.patch('keepercommander.commands.folder.user_choice') as mock_choice:
            mock_choice.return_value = 'n'
            KeeperApiHelper.communicate_expect([pre_delete])
            cmd.execute(params, pattern=[x.name for x in all_folders])
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            mock_choice.return_value = 'y'
            KeeperApiHelper.communicate_expect([pre_delete, 'delete'])
            cmd.execute(params, pattern=[x.name for x in all_folders])
            self.assertTrue(KeeperApiHelper.is_expect_empty())


    def test_move_success(self):
        params = get_synced_params()
        cmd = folder.FolderMoveCommand()

        user_folder = next(iter([x for x in params.folder_cache.values() if x.type == 'user_folder']))
        shared_folder = next(iter([x for x in params.folder_cache.values() if x.type == 'shared_folder']))

        root_record_uid = next(iter(params.subfolder_record_cache['']))
        sf_record_uid = next(iter(params.subfolder_record_cache[shared_folder.uid]))

        KeeperApiHelper.communicate_expect(['move'])
        cmd.execute(params, src=root_record_uid, dst=user_folder.uid)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

        KeeperApiHelper.communicate_expect(['move'])
        cmd.execute(params, src=sf_record_uid, dst=user_folder.name)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

        KeeperApiHelper.communicate_expect(['move'])
        cmd.execute(params, src=root_record_uid, dst=user_folder.uid)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_move_invalid_input(self):
        params = get_synced_params()
        cmd = folder.FolderMoveCommand()

        user_folder = next(iter([x for x in params.folder_cache.values() if x.type == 'user_folder']))
        # shared_folder = next(iter([x for x in params.folder_cache.values() if x.type == 'shared_folder']))

        root_record_uid = next(iter(params.subfolder_record_cache['']))

        with self.assertRaises(CommandError):
            cmd.execute(params, src='Invalid Record', dst=user_folder.uid)
        with self.assertRaises(CommandError):
            cmd.execute(params, src=root_record_uid, dst='Invalid Folder')
        with self.assertRaises(CommandError):
            cmd.execute(params, src=user_folder.uid, dst=user_folder.uid)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: folder.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import record_pb2 as record__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x66older.proto\x12\x06\x46older\x1a\x0crecord.proto\"\\\n\x10\x45ncryptedDataKey\x12\x14\n\x0c\x65ncryptedKey\x18\x01 \x01(\x0c\x12\x32\n\x10\x65ncryptedKeyType\x18\x02 \x01(\x0e\x32\x18.Folder.EncryptedKeyType\"\x82\x01\n\x16SharedFolderRecordData\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x0e\n\x06userId\x18\x03 \x01(\x05\x12\x32\n\x10\x65ncryptedDataKey\x18\x04 \x03(\x0b\x32\x18.Folder.EncryptedDataKey\"\\\n\x1aSharedFolderRecordDataList\x12>\n\x16sharedFolderRecordData\x18\x01 \x03(\x0b\x32\x1e.Folder.SharedFolderRecordData\"_\n\x15SharedFolderRecordFix\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12 \n\x18\x65ncryptedRecordFolderKey\x18\x03 \x01(\x0c\"Y\n\x19SharedFolderRecordFixList\x12<\n\x15sharedFolderRecordFix\x18\x01 \x03(\x0b\x32\x1d.Folder.SharedFolderRecordFix\"\xa2\x02\n\rRecordRequest\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12&\n\nrecordType\x18\x02 \x01(\x0e\x32\x12.Folder.RecordType\x12\x12\n\nrecordData\x18\x03 \x01(\x0c\x12\x1a\n\x12\x65ncryptedRecordKey\x18\x04 \x01(\x0c\x12&\n\nfolderType\x18\x05 \x01(\x0e\x32\x12.Folder.FolderType\x12\x12\n\nhowLongAgo\x18\x06 \x01(\x03\x12\x11\n\tfolderUid\x18\x07 \x01(\x0c\x12 \n\x18\x65ncryptedRecordFolderKey\x18\x08 \x01(\x0c\x12\r\n\x05\x65xtra\x18\t \x01(\x0c\x12\x15\n\rnonSharedData\x18\n \x01(\x0c\x12\x0f\n\x07\x66ileIds\x18\x0b \x03(\x03\"E\n\x0eRecordResponse\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0e\n\x06status\x18\x03 \x01(\t\"\x80\x01\n\x12SharedFolderFields\x12\x1b\n\x13\x65ncryptedFolderName\x18\x01 \x01(\x0c\x12\x13\n\x0bmanageUsers\x18\x02 \x01(\x08\x12\x15\n\rmanageRecords\x18\x03 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x04 \x01(\x08\x12\x10\n\x08\x63\x61nShare\x18\x05 \x01(\x08\"3\n\x18SharedFolderFolderFields\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\"\x8f\x02\n\rFolderRequest\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12&\n\nfolderType\x18\x02 \x01(\x0e\x32\x12.Folder.FolderType\x12\x17\n\x0fparentFolderUid\x18\x03 \x01(\x0c\x12\x12\n\nfolderData\x18\x04 \x01(\x0c\x12\x1a\n\x12\x65ncryptedFolderKey\x18\x05 \x01(\x0c\x12\x36\n\x12sharedFolderFields\x18\x06 \x01(\x0b\x32\x1a.Folder.SharedFolderFields\x12\x42\n\x18sharedFolderFolderFields\x18\x07 \x01(\x0b\x32 .Folder.SharedFolderFolderFields\"E\n\x0e\x46olderResponse\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0e\n\x06status\x18\x03 \x01(\t\"w\n\x19ImportFolderRecordRequest\x12,\n\rfolderRequest\x18\x01 \x03(\x0b\x32\x15.Folder.FolderRequest\x12,\n\rrecordRequest\x18\x02 \x03(\x0b\x32\x15.Folder.RecordRequest\"|\n\x1aImportFolderRecordResponse\x12.\n\x0e\x66olderResponse\x18\x01 \x03(\x0b\x32\x16.Folder.FolderResponse\x12.\n\x0erecordResponse\x18\x02 \x03(\x0b\x32\x16.Folder.RecordResponse\"\xc9\x02\n\x18SharedFolderUpdateRecord\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\x12\x0f\n\x07teamUid\x18\x03 \x01(\x0c\x12(\n\x07\x63\x61nEdit\x18\x04 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12)\n\x08\x63\x61nShare\x18\x05 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x1a\n\x12\x65ncryptedRecordKey\x18\x06 \x01(\x0c\x12\x10\n\x08revision\x18\x07 \x01(\x05\x12\x12\n\nexpiration\x18\x08 \x01(\x12\x12=\n\x15timerNotificationType\x18\t \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x1a\n\x12rotateOnExpiration\x18\n \x01(\x08\"\xcc\x02\n\x16SharedFolderUpdateUser\x12\x10\n\x08username\x18\x01 \x01(\t\x12,\n\x0bmanageUsers\x18\x02 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12.\n\rmanageRecords\x18\x03 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x1b\n\x0fsharedFolderKey\x18\x04 \x01(\x0c\x42\x02\x18\x01\x12\x12\n\nexpiration\x18\x05 \x01(\x12\x12=\n\x15timerNotificationType\x18\x06 \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x36\n\x14typedSharedFolderKey\x18\x07 \x01(\x0b\x32\x18.Folder.EncryptedDataKey\x12\x1a\n\x12rotateOnExpiration\x18\x08 \x01(\x08\"\x99\x02\n\x16SharedFolderUpdateTeam\x12\x0f\n\x07teamUid\x18\x01 \x01(\x0c\x12\x13\n\x0bmanageUsers\x18\x02 \x01(\x08\x12\x15\n\rmanageRecords\x18\x03 \x01(\x08\x12\x1b\n\x0fsharedFolderKey\x18\x04 \x01(\x0c\x42\x02\x18\x01\x12\x12\n\nexpiration\x18\x05 \x01(\x12\x12=\n\x15timerNotificationType\x18\x06 \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x36\n\x14typedSharedFolderKey\x18\x07 \x01(\x0b\x32\x18.Folder.EncryptedDataKey\x12\x1a\n\x12rotateOnExpiration\x18\x08 \x01(\x08\"\x8e\x07\n\x1bSharedFolderUpdateV3Request\x12,\n$sharedFolderUpdateOperation_dont_use\x18\x01 \x01(\x05\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\x12!\n\x19\x65ncryptedSharedFolderName\x18\x03 \x01(\x0c\x12\x10\n\x08revision\x18\x04 \x01(\x03\x12\x13\n\x0b\x66orceUpdate\x18\x05 \x01(\x08\x12\x13\n\x0b\x66romTeamUid\x18\x06 \x01(\x0c\x12\x33\n\x12\x64\x65\x66\x61ultManageUsers\x18\x07 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x35\n\x14\x64\x65\x66\x61ultManageRecords\x18\x08 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12/\n\x0e\x64\x65\x66\x61ultCanEdit\x18\t \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x30\n\x0f\x64\x65\x66\x61ultCanShare\x18\n \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12?\n\x15sharedFolderAddRecord\x18\x0b \x03(\x0b\x32 .Folder.SharedFolderUpdateRecord\x12;\n\x13sharedFolderAddUser\x18\x0c \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateUser\x12;\n\x13sharedFolderAddTeam\x18\r \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateTeam\x12\x42\n\x18sharedFolderUpdateRecord\x18\x0e \x03(\x0b\x32 .Folder.SharedFolderUpdateRecord\x12>\n\x16sharedFolderUpdateUser\x18\x0f \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateUser\x12>\n\x16sharedFolderUpdateTeam\x18\x10 \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateTeam\x12 \n\x18sharedFolderRemoveRecord\x18\x11 \x03(\x0c\x12\x1e\n\x16sharedFolderRemoveUser\x18\x12 \x03(\t\x12\x1e\n\x16sharedFolderRemoveTeam\x18\x13 \x03(\x0c\x12\x19\n\x11sharedFolderOwner\x18\x14 \x01(\t\"c\n\x1dSharedFolderUpdateV3RequestV2\x12\x42\n\x15sharedFoldersUpdateV3\x18\x01 \x03(\x0b\x32#.Folder.SharedFolderUpdateV3Request\"C\n\x1eSharedFolderUpdateRecordStatus\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\"@\n\x1cSharedFolderUpdateUserStatus\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\"?\n\x1cSharedFolderUpdateTeamStatus\x12\x0f\n\x07teamUid\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\"\x88\x06\n\x1cSharedFolderUpdateV3Response\x12\x10\n\x08revision\x18\x01 \x01(\x03\x12K\n\x1bsharedFolderAddRecordStatus\x18\x02 \x03(\x0b\x32&.Folder.SharedFolderUpdateRecordStatus\x12G\n\x19sharedFolderAddUserStatus\x18\x03 \x03(\x0b\x32$.Folder.SharedFolderUpdateUserStatus\x12G\n\x19sharedFolderAddTeamStatus\x18\x04 \x03(\x0b\x32$.Folder.SharedFolderUpdateTeamStatus\x12N\n\x1esharedFolderUpdateRecordStatus\x18\x05 \x03(\x0b\x32&.Folder.SharedFolderUpdateRecordStatus\x12J\n\x1csharedFolderUpdateUserStatus\x18\x06 \x03(\x0b\x32$.Folder.SharedFolderUpdateUserStatus\x12J\n\x1csharedFolderUpdateTeamStatus\x18\x07 \x03(\x0b\x32$.Folder.SharedFolderUpdateTeamStatus\x12N\n\x1esharedFolderRemoveRecordStatus\x18\x08 \x03(\x0b\x32&.Folder.SharedFolderUpdateRecordStatus\x12J\n\x1csharedFolderRemoveUserStatus\x18\t \x03(\x0b\x32$.Folder.SharedFolderUpdateUserStatus\x12J\n\x1csharedFolderRemoveTeamStatus\x18\n \x03(\x0b\x32$.Folder.SharedFolderUpdateTeamStatus\x12\x17\n\x0fsharedFolderUid\x18\x0c \x01(\x0c\x12\x0e\n\x06status\x18\r \x01(\t\"m\n\x1eSharedFolderUpdateV3ResponseV2\x12K\n\x1dsharedFoldersUpdateV3Response\x18\x01 \x03(\x0b\x32$.Folder.SharedFolderUpdateV3Response\"\xfa\x01\n)GetDeletedSharedFoldersAndRecordsResponse\x12\x32\n\rsharedFolders\x18\x01 \x03(\x0b\x32\x1b.Folder.DeletedSharedFolder\x12>\n\x13sharedFolderRecords\x18\x02 \x03(\x0b\x32!.Folder.DeletedSharedFolderRecord\x12\x34\n\x11\x64\x65letedRecordData\x18\x03 \x03(\x0b\x32\x19.Folder.DeletedRecordData\x12#\n\tusernames\x18\x04 \x03(\x0b\x32\x10.Folder.Username\"\xd1\x01\n\x13\x44\x65letedSharedFolder\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x11\n\tfolderUid\x18\x02 \x01(\x0c\x12\x11\n\tparentUid\x18\x03 \x01(\x0c\x12\x17\n\x0fsharedFolderKey\x18\x04 \x01(\x0c\x12-\n\rfolderKeyType\x18\x05 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\x13\n\x0b\x64\x61teDeleted\x18\x07 \x01(\x03\x12\x10\n\x08revision\x18\x08 \x01(\x03\"\x81\x01\n\x19\x44\x65letedSharedFolderRecord\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x17\n\x0fsharedRecordKey\x18\x03 \x01(\x0c\x12\x13\n\x0b\x64\x61teDeleted\x18\x04 \x01(\x03\x12\x10\n\x08revision\x18\x05 \x01(\x03\"\x85\x01\n\x11\x44\x65letedRecordData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08ownerUid\x18\x02 \x01(\x0c\x12\x10\n\x08revision\x18\x03 \x01(\x03\x12\x1a\n\x12\x63lientModifiedTime\x18\x04 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x12\x0f\n\x07version\x18\x06 \x01(\x05\"0\n\x08Username\x12\x12\n\naccountUid\x18\x01 \x01(\x0c\x12\x10\n\x08username\x18\x02 \x01(\t\"\x8a\x01\n,RestoreDeletedSharedFoldersAndRecordsRequest\x12,\n\x07\x66olders\x18\x01 \x03(\x0b\x32\x1b.Folder.RestoreSharedObject\x12,\n\x07records\x18\x02 \x03(\x0b\x32\x1b.Folder.RestoreSharedObject\"<\n\x13RestoreSharedObject\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x12\n\nrecordUids\x18\x02 \x03(\x0c*\x1a\n\nRecordType\x12\x0c\n\x08password\x10\x00*^\n\nFolderType\x12\x12\n\x0e\x64\x65\x66\x61ult_folder\x10\x00\x12\x0f\n\x0buser_folder\x10\x01\x12\x11\n\rshared_folder\x10\x02\x12\x18\n\x14shared_folder_folder\x10\x03*\x96\x01\n\x10\x45ncryptedKeyType\x12\n\n\x06no_key\x10\x00\x12\x19\n\x15\x65ncrypted_by_data_key\x10\x01\x12\x1b\n\x17\x65ncrypted_by_public_key\x10\x02\x12\x1d\n\x19\x65ncrypted_by_data_key_gcm\x10\x03\x12\x1f\n\x1b\x65ncrypted_by_public_key_ecc\x10\x04*M\n\x0fSetBooleanValue\x12\x15\n\x11\x42OOLEAN_NO_CHANGE\x10\x00\x12\x10\n\x0c\x42OOLEAN_TRUE\x10\x01\x12\x11\n\rBOOLEAN_FALSE\x10\x02\x42\"\n\x18\x63om.keepersecurity.protoB\x06\x46olderb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'folder_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\006Folder'
  _globals['_SHAREDFOLDERUPDATEUSER'].fields_by_name['sharedFolderKey']._options = None
  _globals['_SHAREDFOLDERUPDATEUSER'].fields_by_name['sharedFolderKey']._serialized_options = b'\030\001'
  _globals['_SHAREDFOLDERUPDATETEAM'].fields_by_name['sharedFolderKey']._options = None
  _globals['_SHAREDFOLDERUPDATETEAM'].fields_by_name['sharedFolderKey']._serialized_options = b'\030\001'
  _globals['_RECORDTYPE']._serialized_start=5728
  _globals['_RECORDTYPE']._serialized_end=5754
  _globals['_FOLDERTYPE']._serialized_start=5756
  _globals['_FOLDERTYPE']._serialized_end=5850
  _globals['_ENCRYPTEDKEYTYPE']._serialized_start=5853
  _globals['_ENCRYPTEDKEYTYPE']._serialized_end=6003
  _globals['_SETBOOLEANVALUE']._serialized_start=6005
  _globals['_SETBOOLEANVALUE']._serialized_end=6082
  _globals['_ENCRYPTEDDATAKEY']._serialized_start=38
  _globals['_ENCRYPTEDDATAKEY']._serialized_end=130
  _globals['_SHAREDFOLDERRECORDDATA']._serialized_start=133
  _globals['_SHAREDFOLDERRECORDDATA']._serialized_end=263
  _globals['_SHAREDFOLDERRECORDDATALIST']._serialized_start=265
  _globals['_SHAREDFOLDERRECORDDATALIST']._serialized_end=357
  _globals['_SHAREDFOLDERRECORDFIX']._serialized_start=359
  _globals['_SHAREDFOLDERRECORDFIX']._serialized_end=454
  _globals['_SHAREDFOLDERRECORDFIXLIST']._serialized_start=456
  _globals['_SHAREDFOLDERRECORDFIXLIST']._serialized_end=545
  _globals['_RECORDREQUEST']._serialized_start=548
  _globals['_RECORDREQUEST']._serialized_end=838
  _globals['_RECORDRESPONSE']._serialized_start=840
  _globals['_RECORDRESPONSE']._serialized_end=909
  _globals['_SHAREDFOLDERFIELDS']._serialized_start=912
  _globals['_SHAREDFOLDERFIELDS']._serialized_end=1040
  _globals['_SHAREDFOLDERFOLDERFIELDS']._serialized_start=1042
  _globals['_SHAREDFOLDERFOLDERFIELDS']._serialized_end=1093
  _globals['_FOLDERREQUEST']._serialized_start=1096
  _globals['_FOLDERREQUEST']._serialized_end=1367
  _globals['_FOLDERRESPONSE']._serialized_start=1369
  _globals['_FOLDERRESPONSE']._serialized_end=1438
  _globals['_IMPORTFOLDERRECORDREQUEST']._serialized_start=1440
  _globals['_IMPORTFOLDERRECORDREQUEST']._serialized_end=1559
  _globals['_IMPORTFOLDERRECORDRESPONSE']._serialized_start=1561
  _globals['_IMPORTFOLDERRECORDRESPONSE']._serialized_end=1685
  _globals['_SHAREDFOLDERUPDATERECORD']._serialized_start=1688
  _globals['_SHAREDFOLDERUPDATERECORD']._serialized_end=2017
  _globals['_SHAREDFOLDERUPDATEUSER']._serialized_start=2020
  _globals['_SHAREDFOLDERUPDATEUSER']._serialized_end=2352
  _globals['_SHAREDFOLDERUPDATETEAM']._serialized_start=2355
  _globals['_SHAREDFOLDERUPDATETEAM']._serialized_end=2636
  _globals['_SHAREDFOLDERUPDATEV3REQUEST']._serialized_start=2639
  _globals['_SHAREDFOLDERUPDATEV3REQUEST']._serialized_end=3549
  _globals['_SHAREDFOLDERUPDATEV3REQUESTV2']._serialized_start=3551
  _globals['_SHAREDFOLDERUPDATEV3REQUESTV2']._serialized_end=3650
  _globals['_SHAREDFOLDERUPDATERECORDSTATUS']._serialized_start=3652
  _globals['_SHAREDFOLDERUPDATERECORDSTATUS']._serialized_end=3719
  _globals['_SHAREDFOLDERUPDATEUSERSTATUS']._serialized_start=3721
  _globals['_SHAREDFOLDERUPDATEUSERSTATUS']._serialized_end=3785
  _globals['_SHAREDFOLDERUPDATETEAMSTATUS']._serialized_start=3787
  _globals['_SHAREDFOLDERUPDATETEAMSTATUS']._serialized_end=3850
  _globals['_SHAREDFOLDERUPDATEV3RESPONSE']._serialized_start=3853
  _globals['_SHAREDFOLDERUPDATEV3RESPONSE']._serialized_end=4629
  _globals['_SHAREDFOLDERUPDATEV3RESPONSEV2']._serialized_start=4631
  _globals['_SHAREDFOLDERUPDATEV3RESPONSEV2']._serialized_end=4740
  _globals['_GETDELETEDSHAREDFOLDERSANDRECORDSRESPONSE']._serialized_start=4743
  _globals['_GETDELETEDSHAREDFOLDERSANDRECORDSRESPONSE']._serialized_end=4993
  _globals['_DELETEDSHAREDFOLDER']._serialized_start=4996
  _globals['_DELETEDSHAREDFOLDER']._serialized_end=5205
  _globals['_DELETEDSHAREDFOLDERRECORD']._serialized_start=5208
  _globals['_DELETEDSHAREDFOLDERRECORD']._serialized_end=5337
  _globals['_DELETEDRECORDDATA']._serialized_start=5340
  _globals['_DELETEDRECORDDATA']._serialized_end=5473
  _globals['_USERNAME']._serialized_start=5475
  _globals['_USERNAME']._serialized_end=5523
  _globals['_RESTOREDELETEDSHAREDFOLDERSANDRECORDSREQUEST']._serialized_start=5526
  _globals['_RESTOREDELETEDSHAREDFOLDERSANDRECORDSREQUEST']._serialized_end=5664
  _globals['_RESTORESHAREDOBJECT']._serialized_start=5666
  _globals['_RESTORESHAREDOBJECT']._serialized_end=5726
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from pathlib import Path
from datetime import datetime
from typing import Dict, Any
from .cli_handler import CommandHandler
from .config_validation import ConfigValidator
from ..decorators.logging import logger, debug_decorator
from ..util.api_key import generate_api_key
from ..util.exceptions import ValidationError
from ... import utils
from ...params import KeeperParams

class RecordHandler:
    def __init__(self):
        self.validator = ConfigValidator()
        self.cli_handler = CommandHandler()

    @debug_decorator
    def create_record(self, is_advanced_security_enabled: str, commands: str) -> Dict[str, Any]:
        """Create a new configuration record."""
        api_key = generate_api_key()
        record = self._create_base_record(api_key, commands)
        
        if is_advanced_security_enabled == "y":
            logger.debug("Adding expiration to record (advanced security enabled)")
            self._add_expiration_to_record(record)
            
        print(f'Generated API key: {api_key}')
        return record

    def update_or_add_record(self, params: KeeperParams, title: str, config_path: Path) -> None:
        """Update existing record or add new one."""
        try:
            record_uid = self.cli_handler.find_config_record(params, title)
            
            config_path_str = f"'{config_path.as_posix()}'"

            command = (
                f"record-update --force --record {record_uid} "
                f"--title '{title}' --record-type=login f.file={config_path_str}"
            ) if record_uid else (
                f"record-add --title='{title}' "
                f"--record-type=login f.file={config_path_str}"
            )
            if record_uid:
                self.cli_handler.execute_cli_command(params, f"delete-attachment {record_uid} --name service_config.json --name service_config.yaml")
            logger.debug(f"{'Updating' if record_uid else 'Creating'} record")
            
            self.cli_handler.execute_cli_command(params, command)

            if not record_uid:
                self.record_uid = self.cli_handler.find_config_record(params, title)

        except Exception as e:
            print(f"Error updating/adding record: {e}")

    def update_or_add_cert_record(self, params: KeeperParams, title: str) -> None:
        """Update existing certificate record or add a new one in Keeper Vault."""
        try:
            record_uid = self.cli_handler.find_config_record(params, title)

            keeper_dir = utils.get_default_path()
            valid_extensions = {".pem", ".crt", ".cer", ".key"}

            cert_files = [file for file in keeper_dir.glob("*") if file.suffix in valid_extensions]


            if not cert_files:
                raise FileNotFoundError("No certificate files found in .keeper directory.")

            file_args = " ".join(f"f.file='{file.as_posix()}'" for file in cert_files)
            command = (
                f"record-update --force --record {record_uid} "
                f"--title '{title}' --record-type=login {file_args}"
            ) if record_uid else (
                f"record-add --title='{title}' "
                f"--record-type=login {file_args}"
            )

            if record_uid:
                delete_cmd = f"delete-attachment {record_uid} " + " ".join(f"--name {file.name}" for file in cert_files)
                self.cli_handler.execute_cli_command(params, delete_cmd)

            logger.debug(f"{'Updating' if record_uid else 'Creating'} certificate record")

            self.cli_handler.execute_cli_command(params, command)

            if not record_uid:
                self.record_uid = self.cli_handler.find_config_record(params, title)

        except Exception as e:
            logger.debug(f"Error updating/adding certificate record: {e}")

    @debug_decorator
    def _create_base_record(self, api_key: str, commands: str) -> Dict[str, Any]:
        """Create base record structure."""
        return {
            "api-key": api_key,
            "command_list": commands,
            "expiration_timestamp": datetime(9999, 12, 31, 23, 59, 59).isoformat(),
            #"expiration_of_token": ""
        }

    @debug_decorator
    def _add_expiration_to_record(self, record: Dict[str, Any]) -> None:
        """Add expiration details to the record."""
        expiration_str = input(
            "Token Expiration Time (Xm, Xh, Xd) or empty for no expiration: "
        ).strip()

        if not expiration_str:
            #record["expiration_of_token"] = ""
            record["expiration_timestamp"] = datetime(9999, 12, 31, 23, 59, 59).isoformat()
            print("API key set to never expire")
            return

        try:
            expiration_delta = self.validator.parse_expiration_time(expiration_str)
            expiration_time = datetime.now() + expiration_delta
            #record["expiration_of_token"] = expiration_str
            record["expiration_timestamp"] = expiration_time.isoformat()
            print(f"API key will expire at: {record['expiration_timestamp']}")
        except ValidationError as e:
            print(f"Error: {str(e)}")
            self._add_expiration_to_record(record)


from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class BreachWatchInfoType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    RECORD: _ClassVar[BreachWatchInfoType]
    ALTERNATE_PASSWORD: _ClassVar[BreachWatchInfoType]
RECORD: BreachWatchInfoType
ALTERNATE_PASSWORD: BreachWatchInfoType

class BreachWatchRecordRequest(_message.Message):
    __slots__ = ["recordUid", "encryptedData", "breachWatchInfoType", "updateUserWhoScanned"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDATA_FIELD_NUMBER: _ClassVar[int]
    BREACHWATCHINFOTYPE_FIELD_NUMBER: _ClassVar[int]
    UPDATEUSERWHOSCANNED_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    encryptedData: bytes
    breachWatchInfoType: BreachWatchInfoType
    updateUserWhoScanned: bool
    def __init__(self, recordUid: _Optional[bytes] = ..., encryptedData: _Optional[bytes] = ..., breachWatchInfoType: _Optional[_Union[BreachWatchInfoType, str]] = ..., updateUserWhoScanned: bool = ...) -> None: ...

class BreachWatchUpdateRequest(_message.Message):
    __slots__ = ["breachWatchRecordRequest", "encryptedData"]
    BREACHWATCHRECORDREQUEST_FIELD_NUMBER: _ClassVar[int]
    ENCRYPTEDDATA_FIELD_NUMBER: _ClassVar[int]
    breachWatchRecordRequest: _containers.RepeatedCompositeFieldContainer[BreachWatchRecordRequest]
    encryptedData: bytes
    def __init__(self, breachWatchRecordRequest: _Optional[_Iterable[_Union[BreachWatchRecordRequest, _Mapping]]] = ..., encryptedData: _Optional[bytes] = ...) -> None: ...

class BreachWatchRecordStatus(_message.Message):
    __slots__ = ["recordUid", "status", "reason"]
    RECORDUID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    recordUid: bytes
    status: str
    reason: str
    def __init__(self, recordUid: _Optional[bytes] = ..., status: _Optional[str] = ..., reason: _Optional[str] = ...) -> None: ...

class BreachWatchUpdateResponse(_message.Message):
    __slots__ = ["breachWatchRecordStatus"]
    BREACHWATCHRECORDSTATUS_FIELD_NUMBER: _ClassVar[int]
    breachWatchRecordStatus: _containers.RepeatedCompositeFieldContainer[BreachWatchRecordStatus]
    def __init__(self, breachWatchRecordStatus: _Optional[_Iterable[_Union[BreachWatchRecordStatus, _Mapping]]] = ...) -> None: ...

class BreachWatchTokenRequest(_message.Message):
    __slots__ = ["breachWatchToken"]
    BREACHWATCHTOKEN_FIELD_NUMBER: _ClassVar[int]
    breachWatchToken: bytes
    def __init__(self, breachWatchToken: _Optional[bytes] = ...) -> None: ...

class BreachWatchTokenResponse(_message.Message):
    __slots__ = ["breachWatchToken", "clientEncrypted"]
    BREACHWATCHTOKEN_FIELD_NUMBER: _ClassVar[int]
    CLIENTENCRYPTED_FIELD_NUMBER: _ClassVar[int]
    breachWatchToken: bytes
    clientEncrypted: bool
    def __init__(self, breachWatchToken: _Optional[bytes] = ..., clientEncrypted: bool = ...) -> None: ...

class AnonymizedTokenResponse(_message.Message):
    __slots__ = ["domainToken", "emailToken", "passwordToken"]
    DOMAINTOKEN_FIELD_NUMBER: _ClassVar[int]
    EMAILTOKEN_FIELD_NUMBER: _ClassVar[int]
    PASSWORDTOKEN_FIELD_NUMBER: _ClassVar[int]
    domainToken: bytes
    emailToken: bytes
    passwordToken: bytes
    def __init__(self, domainToken: _Optional[bytes] = ..., emailToken: _Optional[bytes] = ..., passwordToken: _Optional[bytes] = ...) -> None: ...

class HashCheck(_message.Message):
    __slots__ = ["hash1", "euid"]
    HASH1_FIELD_NUMBER: _ClassVar[int]
    EUID_FIELD_NUMBER: _ClassVar[int]
    hash1: bytes
    euid: bytes
    def __init__(self, hash1: _Optional[bytes] = ..., euid: _Optional[bytes] = ...) -> None: ...

class BreachWatchStatusRequest(_message.Message):
    __slots__ = ["anonymizedToken", "hashCheck", "removedEuid"]
    ANONYMIZEDTOKEN_FIELD_NUMBER: _ClassVar[int]
    HASHCHECK_FIELD_NUMBER: _ClassVar[int]
    REMOVEDEUID_FIELD_NUMBER: _ClassVar[int]
    anonymizedToken: bytes
    hashCheck: _containers.RepeatedCompositeFieldContainer[HashCheck]
    removedEuid: _containers.RepeatedScalarFieldContainer[bytes]
    def __init__(self, anonymizedToken: _Optional[bytes] = ..., hashCheck: _Optional[_Iterable[_Union[HashCheck, _Mapping]]] = ..., removedEuid: _Optional[_Iterable[bytes]] = ...) -> None: ...

class HashStatus(_message.Message):
    __slots__ = ["hash1", "euid", "breachDetected"]
    HASH1_FIELD_NUMBER: _ClassVar[int]
    EUID_FIELD_NUMBER: _ClassVar[int]
    BREACHDETECTED_FIELD_NUMBER: _ClassVar[int]
    hash1: bytes
    euid: bytes
    breachDetected: bool
    def __init__(self, hash1: _Optional[bytes] = ..., euid: _Optional[bytes] = ..., breachDetected: bool = ...) -> None: ...

class BreachWatchStatusResponse(_message.Message):
    __slots__ = ["hashStatus"]
    HASHSTATUS_FIELD_NUMBER: _ClassVar[int]
    hashStatus: _containers.RepeatedCompositeFieldContainer[HashStatus]
    def __init__(self, hashStatus: _Optional[_Iterable[_Union[HashStatus, _Mapping]]] = ...) -> None: ...

class EnterprisePublicKeyResponse(_message.Message):
    __slots__ = ["enterprisePublicKey", "enterpriseECCPublicKey"]
    ENTERPRISEPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEECCPUBLICKEY_FIELD_NUMBER: _ClassVar[int]
    enterprisePublicKey: bytes
    enterpriseECCPublicKey: bytes
    def __init__(self, enterprisePublicKey: _Optional[bytes] = ..., enterpriseECCPublicKey: _Optional[bytes] = ...) -> None: ...

class FreeScanRequest(_message.Message):
    __slots__ = ["hashedEmail"]
    HASHEDEMAIL_FIELD_NUMBER: _ClassVar[int]
    hashedEmail: bytes
    def __init__(self, hashedEmail: _Optional[bytes] = ...) -> None: ...

class FreeScanResponse(_message.Message):
    __slots__ = ["emailBreaches", "passwordBreaches"]
    EMAILBREACHES_FIELD_NUMBER: _ClassVar[int]
    PASSWORDBREACHES_FIELD_NUMBER: _ClassVar[int]
    emailBreaches: int
    passwordBreaches: int
    def __init__(self, emailBreaches: _Optional[int] = ..., passwordBreaches: _Optional[int] = ...) -> None: ...

class PaidUserRequest(_message.Message):
    __slots__ = ["email"]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    email: str
    def __init__(self, email: _Optional[str] = ...) -> None: ...

class PaidUserResponse(_message.Message):
    __slots__ = ["paidUser"]
    PAIDUSER_FIELD_NUMBER: _ClassVar[int]
    paidUser: bool
    def __init__(self, paidUser: bool = ...) -> None: ...

class DetailedScanRequest(_message.Message):
    __slots__ = ["email"]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    email: str
    def __init__(self, email: _Optional[str] = ...) -> None: ...

class UseOneTimeTokenRequest(_message.Message):
    __slots__ = ["token"]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    token: bytes
    def __init__(self, token: _Optional[bytes] = ...) -> None: ...

class BreachEvent(_message.Message):
    __slots__ = ["site", "email", "passwordInBreach", "date", "description"]
    SITE_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    PASSWORDINBREACH_FIELD_NUMBER: _ClassVar[int]
    DATE_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    site: str
    email: str
    passwordInBreach: bool
    date: str
    description: str
    def __init__(self, site: _Optional[str] = ..., email: _Optional[str] = ..., passwordInBreach: bool = ..., date: _Optional[str] = ..., description: _Optional[str] = ...) -> None: ...

class UseOneTimeTokenResponse(_message.Message):
    __slots__ = ["emailBreaches", "passwordBreaches", "breachEvents", "email"]
    EMAILBREACHES_FIELD_NUMBER: _ClassVar[int]
    PASSWORDBREACHES_FIELD_NUMBER: _ClassVar[int]
    BREACHEVENTS_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    emailBreaches: int
    passwordBreaches: int
    breachEvents: _containers.RepeatedCompositeFieldContainer[BreachEvent]
    email: str
    def __init__(self, emailBreaches: _Optional[int] = ..., passwordBreaches: _Optional[int] = ..., breachEvents: _Optional[_Iterable[_Union[BreachEvent, _Mapping]]] = ..., email: _Optional[str] = ...) -> None: ...

class OneTimeUseToken(_message.Message):
    __slots__ = ["email", "pad"]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    PAD_FIELD_NUMBER: _ClassVar[int]
    email: str
    pad: str
    def __init__(self, email: _Optional[str] = ..., pad: _Optional[str] = ...) -> None: ...

class FreePasswordScanRequest(_message.Message):
    __slots__ = ["hashedPassword"]
    HASHEDPASSWORD_FIELD_NUMBER: _ClassVar[int]
    hashedPassword: bytes
    def __init__(self, hashedPassword: _Optional[bytes] = ...) -> None: ...

class FreePasswordScanResponse(_message.Message):
    __slots__ = ["passwordBreaches"]
    PASSWORDBREACHES_FIELD_NUMBER: _ClassVar[int]
    passwordBreaches: int
    def __init__(self, passwordBreaches: _Optional[int] = ...) -> None: ...



from unittest import TestCase, mock

from data_vault import VaultEnvironment, get_synced_params
from keepercommander.api import sync_down, crypto, utils
from keepercommander.proto import SyncDown_pb2

vault_env = VaultEnvironment()


class TestSyncDown(TestCase):
    def test_full_sync(self):
        params = get_synced_params()

        self.assertEqual(len(params.record_cache), 3)
        self.assertEqual(len(params.shared_folder_cache), 1)
        self.assertEqual(len(params.team_cache), 1)
        self.assert_key_unencrypted(params)

    def test_sync_remove_owned_records(self):
        params = get_synced_params()
        len_before = len(params.record_cache)

        records_to_delete = [x for x, md in params.meta_data_cache.items() if md.get('owner') is True]

        with mock.patch('keepercommander.api.communicate_rest') as mock_comm:
            rs = SyncDown_pb2.SyncDownResponse()
            rs.continuationToken = crypto.get_random_bytes(64)
            rs.removedRecords.extend((utils.base64_url_decode(x) for x in records_to_delete))
            mock_comm.return_value = rs
            sync_down(params)

        self.assertEqual(len(params.record_cache), len_before - len(records_to_delete))
        self.assert_key_unencrypted(params)

    def test_sync_remove_team(self):
        params = get_synced_params()
        teams_to_delete = [x['team_uid'] for x in params.team_cache.values()]

        with mock.patch('keepercommander.api.communicate_rest') as mock_comm:
            rs = SyncDown_pb2.SyncDownResponse()
            rs.continuationToken = crypto.get_random_bytes(64)
            rs.removedTeams.extend((utils.base64_url_decode(x) for x in teams_to_delete))
            mock_comm.return_value = rs
            sync_down(params)

        self.assertEqual(len(params.record_cache), 3)
        self.assertEqual(len(params.team_cache), 0)
        self.assert_key_unencrypted(params)

    def test_sync_remove_shared_folder_then_team(self):
        params = get_synced_params()
        sf_to_delete = [x['shared_folder_uid'] for x in params.shared_folder_cache.values()]

        with mock.patch('keepercommander.api.communicate_rest') as mock_comm:
            rs = SyncDown_pb2.SyncDownResponse()
            rs.continuationToken = crypto.get_random_bytes(64)
            rs.removedSharedFolders.extend((utils.base64_url_decode(x) for x in sf_to_delete))
            mock_comm.return_value = rs
            sync_down(params)

        self.assertEqual(len(params.record_cache), 2)
        self.assertEqual(len(params.shared_folder_cache), 0)
        self.assertEqual(len(params.team_cache), 1)
        self.assert_key_unencrypted(params)

        teams_to_delete = [x['team_uid'] for x in params.team_cache.values()]

        with mock.patch('keepercommander.api.communicate_rest') as mock_comm:
            rs = SyncDown_pb2.SyncDownResponse()
            rs.continuationToken = crypto.get_random_bytes(64)
            rs.removedTeams.extend((utils.base64_url_decode(x) for x in teams_to_delete))
            mock_comm.return_value = rs
            sync_down(params)

        self.assertEqual(len(params.record_cache), 2)
        self.assertEqual(len(params.shared_folder_cache), 0)
        self.assertEqual(len(params.team_cache), 0)
        self.assert_key_unencrypted(params)

    def test_sync_remove_team_shared_folder(self):
        params = get_synced_params()
        teams_to_delete = [x['team_uid'] for x in params.team_cache.values()]
        sf_to_delete = [x['shared_folder_uid'] for x in params.shared_folder_cache.values()]

        with mock.patch('keepercommander.api.communicate_rest') as mock_comm:
            rs = SyncDown_pb2.SyncDownResponse()
            rs.continuationToken = crypto.get_random_bytes(64)
            rs.removedTeams.extend((utils.base64_url_decode(x) for x in teams_to_delete))
            rs.removedSharedFolders.extend((utils.base64_url_decode(x) for x in sf_to_delete))
            mock_comm.return_value = rs
            sync_down(params)

        self.assertEqual(len(params.record_cache), 2)
        self.assertEqual(len(params.shared_folder_cache), 0)
        self.assertEqual(len(params.team_cache), 0)
        self.assert_key_unencrypted(params)

    def assert_key_unencrypted(self, params):
        for r in params.record_cache.values():
            self.assertTrue('record_key_unencrypted' in r)
        for sf in params.shared_folder_cache.values():
            self.assertTrue('shared_folder_key_unencrypted' in sf)
        for t in params.team_cache.values():
            self.assertTrue('team_key_unencrypted' in t)



#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import datetime
import logging
import os
import re
import shlex
import subprocess
import sys
import threading
import time
from collections import OrderedDict
from pathlib import Path
from typing import Union

from prompt_toolkit import PromptSession
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.shortcuts import CompleteStyle

from . import api, display, ttk
from . import versioning
from .autocomplete import CommandCompleter
from .commands import (
    register_commands, register_enterprise_commands, register_msp_commands,
    aliases, commands, command_info, enterprise_commands, msp_commands
)
from .commands.base import dump_report_data, CliCommand, GroupCommand
from .commands.utils import LoginCommand
from .commands import msp
from .constants import OS_WHICH_CMD, KEEPER_PUBLIC_HOSTS
from .error import CommandError, Error
from .params import KeeperParams
from .subfolder import BaseFolderNode

current_command = None  # type: Union[None, CliCommand]
stack = []
register_commands(commands, aliases, command_info)
enterprise_command_info = OrderedDict()
msp_command_info = OrderedDict()
register_enterprise_commands(enterprise_commands, aliases, enterprise_command_info)
register_msp_commands(msp_commands, aliases, msp_command_info)

not_msp_admin_error_msg = 'This command is restricted to Keeper MSP administrators logged in to MSP ' \
                          'Company. \nIf you are an MSP administrator then try to run `switch-to-msp` ' \
                          'command before executing this command.'

command_info['server'] = 'Sets or displays current Keeper region'

logging.getLogger('asyncio').setLevel(logging.WARNING)


def display_command_help(show_enterprise=False, show_shell=False, show_legacy=False):
    from .command_categories import get_command_category, get_category_order
    from .display import bcolors
    
    alias_lookup = {x[1]: x[0] for x in aliases.items()}
    
    # Collect all commands from all sources
    all_commands = {}
    all_commands.update(command_info)
    if show_enterprise:
        all_commands.update(enterprise_command_info)
        all_commands.update(msp_command_info)
    
    # Group commands by category
    categorized_commands = {}
    for cmd, description in all_commands.items():
        category = get_command_category(cmd)
        if category not in categorized_commands:
            categorized_commands[category] = []
        categorized_commands[category].append((cmd, description))
    
    # Define colors for different categories - more variety and visual appeal
    category_colors = {
        'Record Commands': bcolors.OKGREEN,           # Green - primary functionality
        'Sharing Commands': bcolors.OKBLUE,           # Blue - collaboration  
        'Record Type Commands': bcolors.HEADER,       # Purple/Magenta - special types
        'Import and Exporting Data': bcolors.WARNING, # Yellow - data operations
        'Reporting Commands': '\033[96m',             # Cyan - analytics
        'MSP Management Commands': bcolors.HIGHINTENSITYRED, # Bright Red - MSP admin
        'Enterprise Management Commands': '\033[94m',  # Blue - enterprise admin
        'Secrets Manager Commands': '\033[95m',       # Magenta - KSM
        'BreachWatch Commands': bcolors.FAIL,         # Red - security alerts
        'Device Management Commands': '\033[93m',     # Bright Yellow - devices
        'Service Mode REST API': '\033[36m',          # Dark Cyan - services
        'Miscellaneous Commands': '\033[37m',         # Light Gray - utilities
        'KeeperPAM Commands': '\033[92m',            # Bright Green - PAM
        'Legacy Commands': '\033[90m',               # Dark Gray - deprecated
        'Other': bcolors.WHITE
    }

    print(f'\n{bcolors.BOLD}{bcolors.UNDERLINE}Commands:{bcolors.ENDC}')
    print('=' * 80)
    
    # Display commands in category order with colors and separators
    first_category = True
    for category in get_category_order():
        if category not in categorized_commands:
            continue
            
        # Skip Legacy Commands unless specifically requested
        if category == 'Legacy Commands' and not show_legacy:
            continue
            
        # Add separator between categories (except for first one)
        if not first_category:
            print()  # Empty line between categories
        first_category = False
            
        # Sort commands within each category
        commands_in_category = sorted(categorized_commands[category], key=lambda x: x[0])
        
        # Display category header with color
        color = category_colors.get(category, bcolors.WHITE)
        print(f'{color}{bcolors.BOLD}{category}:{bcolors.ENDC}')
        print(f'{color}{"-" * len(category)}{bcolors.ENDC}')
        
        # Special handling for KeeperPAM Commands to show sub-commands
        if category == 'KeeperPAM Commands':
            # Define PAM sub-commands with descriptions
            pam_subcommands = [
                ('pam action', 'Execute action on the Gateway'),
                ('pam config', 'Manage PAM Configurations'),
                ('pam connection', 'Manage Connections'),
                ('pam gateway', 'Manage Gateways'),
                ('pam legacy', 'Switch to legacy PAM commands'),
                ('pam project', 'PAM Project Import/Export'),
                ('pam rbi', 'Manage Remote Browser Isolation'),
                ('pam rotation', 'Manage Rotations'),
                ('pam split', 'Split credentials from legacy PAM Machine'),
                ('pam tunnel', 'Manage Tunnels'),
            ]
            
            # Calculate width for PAM commands
            max_cmd_width = max(len(cmd) for cmd, _ in pam_subcommands)
            
            for cmd_display, description in sorted(pam_subcommands):
                # Bold only the "pam" part
                pam_part = cmd_display.split(' ')[0]  # "pam"
                sub_part = cmd_display.split(' ', 1)[1]  # "action", "config", etc.
                formatted_cmd = f'{bcolors.BOLD}{pam_part}{bcolors.ENDC} {sub_part}'
                # Adjust spacing to account for formatting codes
                spacing = max_cmd_width - len(cmd_display) + len(bcolors.BOLD) + len(bcolors.ENDC)
                print(f'  {formatted_cmd}{" " * spacing}   {description}')
        else:
            # Regular command display for other categories
            max_cmd_width = 0
            cmd_display_list = []
            for cmd, description in commands_in_category:
                alias = alias_lookup.get(cmd) or ''
                alias_str = f' ({alias})' if alias else ''
                cmd_display = f'{cmd}{alias_str}'
                cmd_display_list.append((cmd, alias_str, description))
                max_cmd_width = max(max_cmd_width, len(cmd_display))
            
            # Display commands in this category with proper table alignment
            for cmd, alias_str, description in cmd_display_list:
                cmd_display = f'{cmd}{alias_str}'
                print(f'  {bcolors.BOLD}{cmd_display:<{max_cmd_width}}{bcolors.ENDC}   {description}')

    # Add shell commands if requested
    if show_shell:
        print()  # Separator
        color = bcolors.WHITE
        print(f'{color}{bcolors.BOLD}Shell Commands:{bcolors.ENDC}')
        print(f'{color}{"-" * 14}{bcolors.ENDC}')
        # Calculate max width for shell commands too
        shell_commands = [
            ('clear (c)', 'Clear the screen.'),
            ('history (h)', 'Show command history.'),
            ('shell', 'Use Keeper interactive shell.'),
            ('quit (q)', 'Quit.')
        ]
        shell_max_width = max(len(cmd) for cmd, _ in shell_commands)
        
        for cmd, description in shell_commands:
            print(f'  {bcolors.BOLD}{cmd:<{shell_max_width}}{bcolors.ENDC}   {description}')

    print(f'\n{bcolors.UNDERLINE}Usage:{bcolors.ENDC}')
    print(f"Type '{bcolors.BOLD}help <command>{bcolors.ENDC}' to display help on a specific command")
    if not show_legacy:
        print(f"Type '{bcolors.BOLD}help --legacy{bcolors.ENDC}' to show legacy/deprecated commands")


def is_executing_as_msp_admin():
    return msp.msp_params is not None


def check_if_running_as_mc(params, args):
    if msp.current_mc_id is not None:
        if msp.current_mc_id in msp.mc_params_dict:
            params = msp.mc_params_dict[msp.current_mc_id]
        else:
            msp.current_mc_id = None
    else:                                                       # Not impersonating
        if msp.msp_params is not None:
            params = msp.msp_params
            msp.msp_params = None

    return params, args


def is_enterprise_command(name, command, args):   # type: (str, CliCommand, str) -> bool
    if name in enterprise_commands:
        return True
    elif isinstance(command, GroupCommand):
        args = args.split(' ')
        verb = next(iter(args), None)
        subcommand = command.subcommands.get(verb)
        from .commands.enterprise_common import EnterpriseCommand
        return isinstance(subcommand, EnterpriseCommand)
    else:
        return False


def command_and_args_from_cmd(command_line):
    args = ''
    pos = command_line.find(' ')
    if pos > 0:
        cmd = command_line[:pos]
        args = command_line[pos + 1:].strip()
    else:
        cmd = command_line.strip()

    return cmd, args


def do_command(params, command_line):
    def is_msp(params_local):
        if params_local.enterprise:
            if 'licenses' in params_local.enterprise:
                msp_license = next((x for x in params_local.enterprise['licenses'] if x['lic_status'].startswith('msp')),
                                   None)
                if msp_license:
                    return True
        return False

    if command_line.lower() == 'h' or command_line.lower() == 'history':
        display.formatted_history(stack)
        return

    if command_line.lower().startswith('server'):
        _, sp, server = command_line.partition(' ')
        if server:
            if not params.session_token:
                server = server.strip()
                region = next((x for x in KEEPER_PUBLIC_HOSTS.items()
                               if server.casefold() in [x[0].casefold(), x[1].casefold()]), None)
                if region:
                    params.server = region[1]
                    logging.info('Keeper region is set to %s', region[0])
                else:
                    params.server = server
                    logging.info('Keeper server is set to %s',  params.server)
            else:
                logging.warning('Cannot change Keeper region while logged in')
        else:
            print(params.server)
        return

    if command_line.startswith('ksm'):
        try:
            subprocess.check_call([OS_WHICH_CMD, 'ksm'], stdout=subprocess.DEVNULL)
        except subprocess.CalledProcessError:
            logging.error(
                'Please install the ksm application to run ksm commands.\n'
                'See https://docs.keeper.io/secrets-manager/secrets-manager'
                '/secrets-manager-command-line-interface'
                '#secrets-manager-cli-installation'
            )
        else:
            if sys.platform.startswith('win'):
                subprocess.check_call(command_line)
            else:
                subprocess.check_call(shlex.split(command_line))
        return
    elif '-h' in command_line.lower():
        if command_line.lower().startswith('h ') or command_line.lower().startswith('history '):
            print("usage: history|h [-h]")
            print("\nShow command history.")
            print("\noptional arguments:")
            print("  -h, --help            show this help message and exit")
            return
        elif command_line.lower().startswith('c ') or command_line.lower().startswith('cls ') or command_line.lower().startswith('clear '):
            print("usage: clear|cls|c [-h]")
            print("\nClear the screen.")
            print("\noptional arguments:")
            print("  -h, --help            show this help message and exit")
            return
        elif command_line.lower().startswith('debug '):
            print("usage: debug [-h]")
            print("\nToggle debug mode")
            print("\noptional arguments:")
            print("  -h, --help            show this help message and exit")
            print("  --file=PATH           write DEBUG logs to PATH (does not enable terminal DEBUG)")
            return
        elif command_line.lower().startswith('q ') or command_line.lower().startswith('quit '):
            print("usage: quit|q [-h]")
            print("\nExit commander")
            print("\noptional arguments:")
            print("  -h, --help            show this help message and exit")
            return

    # Track commands history
    if len(stack) == 0 or stack[0] != command_line:
        stack.insert(0, command_line)

    if command_line.lower() == 'c' or command_line.lower() == 'cls' or command_line.lower() == 'clear':
        print(chr(27) + "[2J")

    elif command_line.startswith('debug'):
        try:
            tokens = shlex.split(command_line)
            debug_manager.process_command(tokens, params.batch_mode)
        except Exception as e:
            logging.error(f"Error processing debug command: {e}")


        # Toggle Rust verbose logging if available
        try:
            import keeper_pam_webrtc_rs
            new_debug_state = debug_manager.is_console_debug_on(params.batch_mode)

            level = logging.DEBUG if new_debug_state else logging.INFO
            logging.getLogger('keeper_pam_webrtc_rs').setLevel(level)

            keeper_pam_webrtc_rs.set_verbose_logging(new_debug_state)
            logging.debug('Rust verbose logging %s', 'ON' if new_debug_state else 'OFF')
        except ImportError:
            pass  # Rust library not available, skip
        except Exception as e:
            logging.debug(f'Failed to toggle Rust verbose logging: {e}')

    else:
        cmd, args = command_and_args_from_cmd(command_line)
        if cmd:
            orig_cmd = cmd
            if cmd in aliases and cmd not in commands and cmd not in enterprise_commands and cmd not in msp_commands:
                ali = aliases[cmd]
                if isinstance(ali, (tuple, list)):
                    cmd = ali[0]
                    args = ' '.join(ali[1:]) + ' ' + args
                else:
                    cmd = ali

            if cmd in commands or cmd in enterprise_commands or cmd in msp_commands:
                command = commands.get(cmd) or enterprise_commands.get(cmd) or msp_commands.get(cmd)
                global current_command
                current_command = command

                if command.is_authorised():
                    if not params.session_token:
                        try:
                            LoginCommand().execute(params, email=params.user, password=params.password, new_login=False)
                        except KeyboardInterrupt:
                            logging.info('Canceled')
                            return

                    if is_enterprise_command(cmd, command, args) or cmd in msp_commands:
                        params, args = check_if_running_as_mc(params, args)

                    if is_enterprise_command(cmd, command, args) and not params.enterprise:
                        if is_executing_as_msp_admin():
                            logging.debug("OK to execute command: %s", cmd)
                        else:
                            logging.error('This command is restricted to Keeper Enterprise administrators.')
                            return

                    if cmd in msp_commands:
                        if not is_msp(params):
                            logging.error(not_msp_admin_error_msg)
                            return

                params.event_queue.clear()
                result = command.execute_args(params, args, command=orig_cmd)
                if params.session_token:
                    if params.event_queue:
                        try:
                            rq = {
                                'command': 'audit_event_client_logging',
                                'item_logs': params.event_queue
                            }
                            api.communicate(params, rq)
                        except Exception as e:
                            logging.debug('Post client events error: %s', e)
                        params.event_queue.clear()
                    if params.sync_data:
                        api.sync_down(params)
                return result
            else:
                display_command_help(show_enterprise=(params.enterprise is not None))


def runcommands(params, commands=None, command_delay=0, quiet=False):
    if commands is None:
        commands = params.commands

    keep_running = True
    first_command = True
    timedelay = params.timedelay

    while keep_running:
        for command in commands:
            if first_command:
                first_command = False
            elif command_delay != 0:
                time.sleep(command_delay)

            if not quiet:
                logging.info('Executing [%s]...', command)
            try:
                result = do_command(params, command)
                if result is not None:
                    print(result)
            except CommandError as e:
                msg = f'{e.command}: {e.message}' if e.command else f'{e.message}'
                logging.error(msg)
            except Error as e:
                logging.error("Communication Error: %s", e.message)
            except Exception as e:
                logging.debug(e, exc_info=True)
                logging.error('An unexpected error occurred: %s', sys.exc_info()[0])

        if timedelay == 0:
            keep_running = False
        else:
            logging.info("%s Waiting for %d seconds", datetime.datetime.now().strftime('%Y/%m/%d %H:%M:%S'), timedelay)
            try:
                time.sleep(timedelay)
            except KeyboardInterrupt:
                keep_running = False


def force_quit():
    try:
        if os.name == 'posix':
            subprocess.run('reset')
        elif os.name == 'nt':
            subprocess.run('cls')
        print('Auto-logout timer activated.')
    except:
        pass
    os._exit(0)


prompt_session = None

class DebugManager:
    """Debug manager for console and file logging."""
    
    def __init__(self):
        self.logger = logging.getLogger()
    
    def has_file_logging(self):
        """Check if file logging is active."""
        return any(getattr(h, '_debug_file', False) for h in self.logger.handlers)
    
    def is_console_debug_on(self, batch_mode):
        """Check if console debug is enabled."""
        for h in self.logger.handlers:
            if isinstance(h, logging.StreamHandler) and not getattr(h, '_debug_file', False):
                return h.level == logging.DEBUG
        return self.logger.level == logging.DEBUG and not self.has_file_logging()
    
    def set_console_debug(self, enabled, batch_mode):
        """Set console debug level."""
        level = logging.DEBUG if enabled else (logging.WARNING if batch_mode else logging.INFO)
        for h in self.logger.handlers:
            if isinstance(h, logging.StreamHandler) and not getattr(h, '_debug_file', False):
                h.setLevel(level)
    
    def setup_file_logging(self, file_path):
        """Setup debug file logging."""
        try:
            validated_path = self._validate_log_file_path(file_path)
            
            log_dir = os.path.dirname(validated_path)
            os.makedirs(log_dir, mode=0o750, exist_ok=True)
            
            for h in list(self.logger.handlers):
                if getattr(h, '_debug_file', False):
                    self.logger.removeHandler(h)
                    try:
                        h.close()
                    except (OSError, IOError) as close_error:
                        logging.warning(f'Failed to close log handler: {close_error}')
                    except Exception as unexpected_error:
                        logging.error(f'Unexpected error closing log handler: {unexpected_error}')
            
            # Add new file handler
            fh = logging.FileHandler(validated_path, mode='a', encoding='utf-8')
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s'))
            fh.addFilter(lambda record: record.levelno != logging.INFO)  # Filter out INFO
            fh._debug_file = True
            self.logger.addHandler(fh)
            self.logger.setLevel(logging.DEBUG)
            
            logging.info(f'Debug file logging enabled: {validated_path}')
            return True
        except (ValueError, OSError, IOError) as e:
            logging.error(f'Failed to setup file logging: {e}')
            return False
        except Exception as e:
            logging.error(f'Unexpected error setting up file logging: {e}')
            return False

    def _validate_log_file_path(self, file_path):
        """Validate and sanitize log file path to prevent security issues."""
        if not file_path or not isinstance(file_path, str):
            raise ValueError("File path must be a non-empty string")
        
        sanitized_path = ''.join(char for char in file_path if ord(char) >= 32 and char != '\x7f')
        
        if not sanitized_path:
            raise ValueError("File path contains only invalid characters")
        
        try:
            path_obj = Path(sanitized_path)
            
            resolved_path = path_obj.resolve()
                        
            resolved_str = str(resolved_path).lower()
            forbidden_paths = ['/etc/', '/bin/', '/sbin/', '/usr/bin/', '/usr/sbin/', 
                             '/boot/', '/dev/', '/proc/', '/sys/', '/root/']
            
            for forbidden in forbidden_paths:
                if resolved_str.startswith(forbidden):
                    raise ValueError(f"Access to system directory '{forbidden}' is not allowed")
            
            filename = path_obj.name
            if not filename or filename in ('.', '..'):
                raise ValueError("Invalid filename")
            
            suspicious_patterns = ['..', '~/', '$', '`', ';', '|', '&', '<', '>', '*', '?']
            for pattern in suspicious_patterns:
                if pattern in filename:
                    raise ValueError(f"Filename contains suspicious pattern: '{pattern}'")
            
            valid_extensions = ['.log', '.txt', '.out']
            if not any(filename.lower().endswith(ext) for ext in valid_extensions):
                logging.warning(f"Log file '{filename}' does not have a standard log extension")
            
            return str(resolved_path)
            
        except (OSError, RuntimeError) as e:
            raise ValueError(f"Invalid file path: {e}")
        except Exception as e:
            raise ValueError(f"Path validation failed: {e}")
    
    def _looks_like_filename(self, token):
        """Check if a token looks like a filename with proper validation."""
        if not token or not isinstance(token, str):
            return False
        
        token = token.strip()
        
        if len(token) < 1:
            return False
        
        has_extension = re.search(r'\.[a-zA-Z0-9]{1,10}$', token)
        
        has_path_separator = '/' in token or '\\' in token
        
        looks_like_name = len(token) > 2 and re.match(r'^[a-zA-Z0-9._-]+$', token)
        
        return bool(has_extension or has_path_separator or looks_like_name)
    
    def toggle_console_debug(self, batch_mode):
        """Toggle console debug on/off."""
        current = self.is_console_debug_on(batch_mode)
        new_state = not current
        
        self.set_console_debug(new_state, batch_mode)
        
        if not self.has_file_logging():
            level = logging.DEBUG if new_state else (logging.WARNING if batch_mode else logging.INFO)
            self.logger.setLevel(level)
        
        logging.info(f'Debug {"ON" if new_state else "OFF"}')
        return new_state
    
    def process_command(self, tokens, batch_mode):
        """Process debug command."""
        # Look for --file argument
        file_path = None
        file_flag_present = False
        
        for i, token in enumerate(tokens[1:], 1):
            if token == '--file':
                file_flag_present = True
                if i + 1 < len(tokens):
                    next_token = tokens[i + 1]
                    if not next_token.startswith('-'):
                        file_path = next_token
                break
            elif token.startswith('--file='):
                file_flag_present = True
                file_path = token.split('=', 1)[1]
                # Handle empty value after equals sign
                if not file_path.strip():
                    file_path = None
                break
        
        if file_flag_present and not file_path:
            print("Please specify the file path for logging to file: debug --file <file_path>")
            return False
        elif file_path:
            return self.setup_file_logging(file_path)
        else:
            # No --file flag present, check for potential filename arguments
            if len(tokens) > 1:
                for token in tokens[1:]:
                    # Check if token looks like a filename
                    if not token.startswith('-') and self._looks_like_filename(token):
                        print(f"Please specify the --file flag for logging to file: debug --file {token}")
                        return False
            
            self.toggle_console_debug(batch_mode)
            return True

debug_manager = DebugManager()


def read_command_with_continuation(prompt_session, params):
    """Read command with support for line continuation using backslash."""
    command_lines = []
    continuation_prompt = "... "
    current_prompt = get_prompt(params)
    
    while True:
        if prompt_session is not None:
            line = prompt_session.prompt(current_prompt)
        else:
            line = input(current_prompt)
        
        # Check if line ends with backslash (line continuation)
        # First strip all trailing whitespace, then check for backslash
        stripped_line = line.rstrip()
        if stripped_line.endswith('\\'):
            # Remove the backslash and any remaining whitespace
            line_content = stripped_line[:-1].strip()
            if line_content:  # Only add non-empty lines
                command_lines.append(line_content)
            current_prompt = continuation_prompt
        else:
            # No continuation, add the final line if it has content
            line_content = stripped_line
            if line_content:
                command_lines.append(line_content)
            break
    
    # Join all lines with spaces, ensuring no extra spaces
    # Also clean up any multiple spaces that might have been introduced
    result = ' '.join(command_lines)
    # Replace multiple spaces with single spaces to handle any remaining formatting issues
    import re
    result = re.sub(r'\s+', ' ', result).strip()
    return result


def loop(params):  # type: (KeeperParams) -> int
    global prompt_session
    error_no = 0
    suppress_errno = False

    logging.getLogger().setLevel(logging.DEBUG if params.debug else logging.WARNING if params.batch_mode else logging.INFO)
    enforcement_checked = set()
    prompt_session = None
    if not params.batch_mode:
        if os.isatty(0) and os.isatty(1):
            completer = CommandCompleter(params, aliases)
            prompt_session = PromptSession(multiline=False,
                                           editing_mode=EditingMode.VI,
                                           completer=completer,
                                           complete_style=CompleteStyle.MULTI_COLUMN,
                                           complete_while_typing=False)

        display.welcome()
        versioning.welcome_print_version(params)

    if not params.batch_mode:
        if params.user:
            try:
                LoginCommand().execute(params, email=params.user, password=params.password, new_login=False)
            except KeyboardInterrupt:
                print('')
            except EOFError:
                return 0
            except Exception as e:
                logging.error(e)
        else:
            if params.device_token:
                logging.info('Current Keeper region: %s', params.server)
            else:
                logging.info('Use "server" command to change Keeper region > "server US"')
                for region in KEEPER_PUBLIC_HOSTS:
                    logging.info('\t%s: %s', region, KEEPER_PUBLIC_HOSTS[region])
            logging.info('To login type: login <email>')

    while True:
        if params.session_token:
            ttk.TTK.update(params)

        command = ''
        if len(params.commands) > 0:
            command = params.commands[0].strip()
            params.commands = params.commands[1:]

        try:
            if not command:
                tmer = None
                try:
                    if params.session_token and params.logout_timer > 0:
                        tmer = threading.Timer(params.logout_timer * 60, force_quit)
                        tmer.start()
                    if prompt_session is not None:
                        if params.enforcements and params.user not in enforcement_checked:
                            enforcement_checked.add(params.user)
                            do_command(params, 'check-enforcements')

                        command = read_command_with_continuation(prompt_session, params)
                    else:
                        command = read_command_with_continuation(None, params)
                    if tmer:
                        tmer.cancel()
                        tmer = None
                finally:
                    if tmer:
                        tmer.cancel()

            if command.lower() == 'q' or command.lower() == "quit":
                break

            suppress_errno = False
            command = command.strip()
            if command.startswith("@"):
                suppress_errno = True
                command = command[1:]
            if params.batch_mode:
                logging.info('> %s', command)
            error_no = 1
            result = do_command(params, command)
            error_no = 0
            if result:
                print(result)
        except EOFError:
            break
        except KeyboardInterrupt:
            pass
        except CommandError as e:
            if e.command:
                logging.warning('%s: %s', e.command, e.message)
            else:
                logging.warning('%s', e.message)
        except Error as e:
            logging.error("Communication Error: %s", e.message)
        except Exception as e:
            logging.debug(e, exc_info=True)
            logging.error('An unexpected error occurred: %s. Type "debug" to toggle verbose error output', e)
        finally:
            global current_command
            try:
                if current_command:
                    current_command.clean_up()
            finally:
                current_command = None

        if params.batch_mode and error_no != 0 and not suppress_errno:
            break

    if not params.batch_mode:
        logging.info('\nGoodbye.\n')

    return error_no


def get_prompt(params):
    if params.batch_mode:
        return ''

    if params.session_token:
        if params.current_folder is None:
            if params.root_folder:
                params.current_folder = ''
            else:
                return 'Keeper> '
    else:
        return 'Not logged in> '

    prompt = ''
    f = params.folder_cache[params.current_folder] if params.current_folder in params.folder_cache else params.root_folder
    while True:
        if len(prompt) > 0:
            prompt = '/' + prompt
        name = f.name
        prompt = name + prompt

        if f == params.root_folder:
            break

        if f.parent_uid is not None:
            f = params.folder_cache[f.parent_uid]
        else:
            if f.type == BaseFolderNode.SharedFolderFolderType:
                f = params.folder_cache[f.shared_folder_uid]
            else:
                f = params.root_folder
    if len(prompt) > 40:
        prompt = '...' + prompt[-37:]

    return prompt + '> '



import base64
import hashlib
import re
from http import HTTPStatus
from http.server import HTTPServer, BaseHTTPRequestHandler
from os import environ
from time import sleep
from urllib.parse import urlparse, parse_qs
import logging
import uuid
import requests
import threading
import webbrowser
from prompt_toolkit import HTML, print_formatted_text, prompt
from tabulate import tabulate


from ..importer import BaseImporter, Record, RecordField
import secrets
import string


class CyberArkPortalImporter(BaseImporter):
    """CyberArk Portal Importer
    This importer enables users to import Applications and SecuredItems from the CyberArk User Portal.
    If the User is authenticated via SSO, it will use OAuth2 with PKCE for authorization.
    Otherwise it will attempt to authenticate using the CyberArk Identity API.
    After successful authentication, it creates a Keeper record for each Application,
    and Secured Item in the user's portal.

    The API documentation can be found at
    https://api-docs.cyberark.com/docs/identity-api-reference/

    Yields:
        Keeper login type records for Applications and Passwords and secure note records for SecuredItems.
    """

    LOOP_DELAY = 0.025  # Use quarter millisecond delay between requests to avoid hitting the API rate limits
    TIMEOUT = 10  # Wait up to 10 seconds for CyberArk API requests

    @staticmethod
    def get_url(id_tenant, endpoint):
        return f'https://{id_tenant}.id.cyberark.cloud/{endpoint.rstrip("/")}'

    def do_import(self, filename, **kwargs):
        id_tenant = re.search(r"^([A-Za-z0-9-]+)(\.id\.cyberark\.cloud)?$", filename.removeprefix("https://"))[0]
        username = environ.get("KEEPER_CYBERARK_USERNAME") or prompt("CyberArk User Portal username: ")
        response = requests.post(
            self.get_url(id_tenant, "/Security/StartAuthentication"),
            json={
                "TenantId": id_tenant,
                "Version": "1.0",
                "User": username,
            },
            timeout=self.TIMEOUT,
        )

        if response.status_code != HTTPStatus.OK:
            logging.error(f"Error starting authentication: {response.text}")
            return
        start_auth_result = response.json().get("Result")
        logging.debug(f"Authentication Result: {start_auth_result}")
        redirect = start_auth_result.get("PodFqdn")

        if redirect:
            print_formatted_text(
                HTML(
                    f"Use <ansigreen><i>{redirect.removesuffix('.id.cyberark.cloud')}</i></ansigreen> instead of <ansired>{id_tenant}</ansired> for user <i>{username}</i>."
                )
            )
            return

        if start_auth_result.get("IdpRedirectUrl"):

            class OAuth2Callback:
                """OAuth2 authorization callback
                Handles the callback for SSO users that trigger the authentication flow based on their username.
                CyberArk will redirect the user to it with an authorization code after successful authentication.
                The authorization code is then exchanged for an access token.
                If Commander is running on a remote machine then the callback will not be received.
                In that case, the user can manually enter the authorization code.
                It will be the "code" query parameter of the URL in the address bar of the local browser.
                """

                CLIENT_ID = environ.get("KEEPER_CYBERARK_OAUTH2_CLIENT_ID", str(uuid.uuid4()))
                CODE_VERIFIER = "".join(
                    secrets.choice(string.ascii_letters + string.digits + "-._~") for _ in range(64)
                )
                HTTP_PORT = 38389
                REDIRECT_URI = environ.get("KEEPER_CYBERARK_OAUTH2_REDIRECT_URI", f"http://localhost:{HTTP_PORT}")
                TIMEOUT = 30  # Wait up to 30 seconds for an OAuth2 authorization code callback

                @classmethod
                def get_challenge(cls):
                    """
                    Generate a code challenge based on the code_verifier for PKCE.
                    """
                    return (
                        base64.urlsafe_b64encode(hashlib.sha256(cls.CODE_VERIFIER.encode()).digest())
                        .decode()
                        .rstrip("=")
                    )

                class Handler(BaseHTTPRequestHandler):
                    CODE_HOLDER = {"code": None}  # thread-safe storage for OAuth2 authorization code

                    def fail(self, message, text):
                        self.send_error(400, message, explain=text)

                    def success(self):
                        self.send_response(200)
                        self.send_header("Content-type", "text/html")
                        self.end_headers()
                        self.wfile.write(
                            b"""<html>
                                <body>
                                    <h2>Authentication successful.</h2>
                                    <p>This window will close in one minute or you may close it now.</p>
                                    <script>setTimeout(function() { window.close(); }, 60000);</script>
                                </body>
                            </html>"""
                        )

                    def do_GET(self):
                        query_params = parse_qs(urlparse(self.path).query)
                        if "code" in query_params:
                            self.CODE_HOLDER["code"] = query_params["code"][0]
                            self.success()
                        elif "error" in query_params and "error_description" in query_params:
                            self.fail(query_params["error"][0], query_params["error_description"][0])
                        else:
                            self.fail(
                                "Missing authorization code", "There is no &quot;code&quot; query parameter in the URL."
                            )

                    def log_message(self, format, *args):
                        logging.debug(format % args)

                @classmethod
                def get_code(cls):
                    return OAuth2Callback.Handler.CODE_HOLDER["code"]

                @classmethod
                def set_code(cls, code):
                    OAuth2Callback.Handler.CODE_HOLDER["code"] = code

                @classmethod
                def handle_callback(cls):
                    HTTPServer(("localhost", cls.HTTP_PORT), cls.Handler).handle_request()

                def __init__(self):
                    self.thread = threading.Thread(target=self.handle_callback)

                def start(self):
                    self.thread.start()

                def complete(self):
                    """
                    Wait for the OAuth2 code callback to come via HTTP GET with the code query parameter.
                    If the callback times out, either because Commander is running on a remote machine
                    or the user closed the browser window, then prompt the user to enter the code or quit.
                    """
                    self.thread.join(timeout=self.TIMEOUT)
                    if self.thread.is_alive():
                        logging.warning(
                            f"Timed out waiting for the OAuth2 code callback on {self.REDIRECT_URI} after {self.TIMEOUT} seconds."
                        )
                        self.set_code(
                            prompt("Enter the OAuth2 authorization code or ctrl+c to cancel: ", is_password=True)
                        )

            # Start listening for the callback then initiate the OAuth2 authorization request
            callback = OAuth2Callback()
            callback.start()
            response = requests.post(
                self.get_url(id_tenant, "/OAuth2/Authorize/KeeperCommander"),
                data={
                    "response_type": "code",
                    "redirect_uri": OAuth2Callback.REDIRECT_URI,
                    "client_id": OAuth2Callback.CLIENT_ID,
                    "code_challenge": OAuth2Callback.get_challenge(),
                    "code_challenge_method": "S256",
                    "scope": "UPData",
                },
                timeout=CyberArkPortalImporter.TIMEOUT,
                allow_redirects=False,
            )

            if response.status_code == 302:
                logging.info(f"Opening {response.next.url}")
                webbrowser.open(response.next.url)
                callback.complete()

            # The user did not quit so the code has been received or entered by the user
            response = requests.post(
                self.get_url(id_tenant, "/OAuth2/Token/KeeperCommander"),
                data={
                    "grant_type": "authorization_code",
                    "redirect_uri": OAuth2Callback.REDIRECT_URI,
                    "client_id": OAuth2Callback.CLIENT_ID,
                    "code_verifier": OAuth2Callback.CODE_VERIFIER,
                    "code": callback.get_code(),
                },
                timeout=CyberArkPortalImporter.TIMEOUT,
            )

            if response.status_code != HTTPStatus.OK:
                logging.error(
                    f"HTTP {HTTPStatus(response.status_code).phrase} error exchanging OAuth2 code for token: {response.text}"
                )
                return
            token_data = response.json()
            authentication_token = token_data["access_token"]  # Save the token before munging it!

            if "access_token" not in token_data:
                logging.error(f"Unexpected response from token exchange: {token_data}")
                return
            else:
                # Obfuscate or mask the access_token in the output
                if len(authentication_token) > 12:
                    token_data["access_token"] = authentication_token[:6] + "..." + authentication_token[-6:]
                else:
                    token_data["access_token"] = "***"
                logging.debug(f"Token Data: {token_data}")
        else:
            # Perform authentication using the CyberArk Identity API starting with password
            password_operation = {
                "MechanismId": start_auth_result["Challenges"][0]["Mechanisms"][0]["MechanismId"],
                "Action": "Answer",
                "Answer": environ.get("KEEPER_CYBERARK_PASSWORD")
                or prompt("CyberArk Identity Portal password: ", is_password=True),
            }
            session_id = start_auth_result.get("SessionId")
            advance_auth_request = {
                "TenantId": id_tenant,
                "SessionId": session_id,
            }

            if len(start_auth_result["Challenges"]) > 1:
                challenge_mechs = list(start_auth_result["Challenges"][1]["Mechanisms"])
                # Use a batch-operation to send the password and start OOB authentication in one request
                advance_auth_request["MultipleOperations"] = [
                    password_operation,
                    {
                        "MechanismId": challenge_mechs[0]["MechanismId"],
                        "Action": "StartOOB",
                    },
                ]
            else:
                challenge_mechs = []
                advance_auth_request |= password_operation

            logging.debug(f"Advance Authentication Request: {advance_auth_request}")
            response = requests.post(
                self.get_url(id_tenant, "/Security/AdvanceAuthentication"),
                json=advance_auth_request,
                timeout=self.TIMEOUT,
            )

            if response.status_code != HTTPStatus.OK:
                logging.error(
                    f"HTTP {HTTPStatus(response.status_code).phrase} error advancing authentication: {response.text}"
                )
                return
            advance_auth_result = response.json().get("Result")

            if not advance_auth_result.get("Summary") in ["OobPending", "LoginSuccess"]:
                logging.error(
                    f"Unexpected authentication summary: {advance_auth_result.get('Summary')}, response: {response.text}"
                )
                return
            logging.debug(f"Advance Authentication Result: {advance_auth_result}")

            # Iterate through the challenge mechanisms until we get a successful login or run out of mechanisms
            while challenge_mechs and advance_auth_result.get("Summary") == "OobPending":
                response = requests.post(
                    self.get_url(id_tenant, "/Security/AdvanceAuthentication"),
                    json={
                        "TenantId": id_tenant,
                        "SessionId": session_id,
                        "MechanismId": challenge_mechs[0]["MechanismId"],
                        "Action": "Answer",
                        "Answer": prompt(f"Authentication code from {challenge_mechs[0]['PromptSelectMech']}: "),
                    },
                    timeout=self.TIMEOUT,
                )

                if response.status_code != HTTPStatus.OK:
                    logging.error(
                        f"HTTP {HTTPStatus(response.status_code).phrase} error advancing authentication: {response.text}"
                    )
                    return
                advance_auth_result = response.json().get("Result")

                if advance_auth_result.get("Summary") != "OobPending":
                    challenge_mechs = challenge_mechs[1:]
                elif advance_auth_result.get("Summary") == "LoginSuccess":
                    break

            if advance_auth_result.get("Summary") != "LoginSuccess":
                logging.error(
                    f"Unexpected authentication summary: {advance_auth_result.get('Summary')}, response: {response.text}"
                )
                return
            authentication_token = advance_auth_result.get("Token")

        print_formatted_text(HTML("Authentication <ansigreen>successful</ansigreen>"))

        # Get all the application data (except the password, of course) in one API call
        response = requests.post(
            self.get_url(id_tenant, "/UPRest/GetUPData"),
            headers={"Authorization": f"Bearer {authentication_token}"},
            json={},
            timeout=self.TIMEOUT,
        )

        if response.status_code != HTTPStatus.OK:
            logging.error(f"HTTP {HTTPStatus(response.status_code).phrase} error getting UP data: {response.text}")
            return
        apps = response.json()["Result"]["Apps"]

        if len(apps) > 0:
            print_formatted_text(
                HTML(f"Importing <b>{len(apps)}</b> Applications:\n"),
                tabulate(
                    [{"Application": app["Name"], "Username": app.get("Username", "")} for app in apps],
                    headers="keys",
                    maxcolwidths=[48, 32],
                ),
                end="\n\n",
            )

            for app in apps:
                sleep(self.LOOP_DELAY)
                app_key = app["AppKey"]
                record = Record()
                record.title = app["Name"]
                record.type = "login"
                record.login = app.get("Username", "")
                record.login_url = app.get("Url", "")
                record.notes = app.get("Notes", "")

                if app.get("IsTotpSet"):
                    record.notes += "The CyberArk Application had a TOTP that Keeper could not access to import."

                if app.get("Tags"):
                    record.fields.append(
                        RecordField(type="text", label="Tags", value=", ".join(str(tag) for tag in app["Tags"]))
                    )

                response = requests.post(
                    self.get_url(id_tenant, f"/UPRest/GetMCFA?appkey={app_key}"),
                    headers={"Authorization": f"Bearer {authentication_token}"},
                    json={},
                    timeout=self.TIMEOUT,
                )

                if response.status_code != HTTPStatus.OK:
                    logging.error(
                        f"HTTP {response.status_code} ({HTTPStatus(response.status_code).phrase}) error getting password for app {app_key}: {response.text}"
                    )
                    continue
                elif not response.json().get("Result"):
                    logging.warning(f"No password found for app {app_key}; response: {response.text}")
                else:
                    record.password = response.json()["Result"].get("p", "")

                yield record

        response = requests.post(
            self.get_url(id_tenant, "/UPRest/GetSecuredItemsData"),
            headers={"Authorization": f"Bearer {authentication_token}"},
            json={},
            timeout=self.TIMEOUT,
        )

        if response.status_code != HTTPStatus.OK:
            logging.error(
                f"HTTP {HTTPStatus(response.status_code).phrase} error getting secured items data: {response.text}"
            )
            return
        secured_items = response.json()["Result"]["SecuredItems"]

        if len(secured_items) > 0:
            print_formatted_text(
                HTML(f"Importing <b>{len(secured_items)}</b> Secured Items:\n"),
                tabulate([{"Name": item["Name"]} for item in secured_items], headers="keys"),
                end="\n\n",
            )

            for item in secured_items:
                sleep(self.LOOP_DELAY)
                item_key = item["ItemKey"]
                record = Record()
                record.title = item["Name"]

                if item.get("Tags"):
                    record.fields.append(
                        RecordField(type="text", label="Tags", value=", ".join(str(tag) for tag in item["Tags"]))
                    )

                response = requests.post(
                    self.get_url(id_tenant, f"/UPRest/GetCredsForSecuredItem?sItemKey={item_key}"),
                    headers={"Authorization": f"Bearer {authentication_token}"},
                    json={},
                    timeout=self.TIMEOUT,
                )

                if response.status_code != HTTPStatus.OK:
                    logging.error(
                        f"HTTP {response.status_code} ({HTTPStatus(response.status_code).phrase}) error getting notes for item {item_key}: {response.text}"
                    )
                    continue
                elif not response.json().get("Result"):
                    logging.warning(f"No notes found for secured item {item_key}; response: {response.text}")
                else:
                    itemData = response.json()["Result"]
                    if item["SecuredItemType"] == "Password":
                        record.type = "login"
                        if item.get("IsUsernameSet", False):
                            record.login = itemData["u"]
                        if item.get("IsPasswordSet", False):
                            record.password = itemData["p"]
                    else:
                        record.type = "encryptedNotes"
                    record.fields.append(RecordField(type="note", value=itemData["n"]))

                yield record

        print_formatted_text(HTML("Import <ansigreen>complete</ansigreen>"))



from __future__ import annotations
import logging
from ..__version__ import __version__
from ..exceptions import DAGException, DAGConnectionException
from ..crypto import generate_random_bytes
from ..types import SyncData, SyncQuery, DataPayload
import json
import base64
from typing import Optional, Union, TYPE_CHECKING
if TYPE_CHECKING:  # pragma: no cover
    Logger = Union[logging.RootLogger, logging.Logger]


class ConnectionBase:

    def __init__(self, is_device: bool, logger: Optional[Logger] = None):
        # device is a gateway device if is_device is False then we use user authentication flow
        self.is_device = is_device

        if logger is None:
            logger = logging.getLogger()
        self.logger = logger

    @staticmethod
    def get_record_uid(record: object) -> str:
        pass

    @staticmethod
    def get_key_bytes(record: object) -> bytes:
        pass

    def rest_call_to_router(self, http_method, endpoint, agent, payload_json=None) -> str:
        return ""

    def _endpoint(self, action: str) -> str:
        if action.startswith("/") is False:
            action = "/" + action

        base = "/api/device"
        if not self.is_device:
            base = "/api/user"
        return base + action

    def add_data(self, payload: Union[DataPayload, str], agent: Optional[str] = None):

        # if payload is DataPayload
        if isinstance(payload, DataPayload):
            payload_data = payload.model_dump_json()
        elif isinstance(payload, str):
            payload_data = payload

            # make sure it is a valid json and raise and exception if not. make an exception for the case of a string
            # that is a valid json
            if not payload_data.startswith('{') and not payload_data.endswith('}'):
                raise DAGException(f'Invalid payload: {payload_data}')

            # double check if it is a valid json inside the string
            json.loads(payload_data)
        else:
            raise DAGException(f'Unsupported payload type: {type(payload)}')

        if agent is None:
            f"keeper-dag/{__version__}"

        try:
            self.rest_call_to_router(http_method="POST",
                                     endpoint=self._endpoint("/add_data"),
                                     payload_json=payload_data,
                                     agent=agent)
        except DAGConnectionException as err:
            raise err
        except Exception as err:
            raise DAGException(f"Could not create a new DAG structure: {err}")

    def sync(self, stream_id: str, agent: Optional[str] = None, sync_point: Optional[int] = 0,
             graph_id: Optional[int] = 0) -> SyncData:

        if agent is None:
            f"keeper-dag/{__version__}"

        try:
            sync_query = SyncQuery(
                streamId=stream_id,
                deviceId=base64.urlsafe_b64encode(generate_random_bytes(16)).decode(),
                syncPoint=sync_point,
                graphId=graph_id
            )
            sync_query_json_str = sync_query.model_dump_json()

            data_resp = self.rest_call_to_router(http_method="POST",
                                                 endpoint=self._endpoint("/sync"),
                                                 agent=agent,
                                                 payload_json=sync_query_json_str)
            sync_data_resp = SyncData.model_validate_json(data_resp)

            return sync_data_resp
        except DAGConnectionException as err:
            raise err
        except Exception as err:
            raise DAGException(f"Could not load the DAG structure: {err}")

    def debug_dump(self) -> str:
        return "Connection does not allow debug dump."



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from typing import Dict, Any, Optional
import time
import os
import psutil
    
from ..decorators.logging import logger, debug_decorator
from .service_config import ServiceConfig
from ..util.tunneling import generate_cloudflare_url
from ..util.exceptions import ValidationError

class CloudflareConfigurator:
    # Constants
    _DEFAULT_HEALTH_CHECK_TIMEOUT = 10
    _HEALTH_CHECK_RETRY_INTERVAL = 1
    
    # Tunnel status patterns
    _SUCCESS_PATTERNS = [
        "tunnel started",
        "Connection established", 
        "Tunnel connection established",
        "Registered tunnel connection",
        "Updated to new configuration",
        "trycloudflare.com",
        "cfargotunnel.com"
    ]
    
    _FAILURE_PATTERNS = [
        "failed to dial to edge with quic: timeout",
        "timeout: no recent network activity",
        "failed to dial a quic connection", 
        "connection timeout",
        "network unreachable"
    ]
    
    _SUCCESS_TIMEOUT_MESSAGE = "Cloudflare tunnel status could not be determined within timeout period."
    _NETWORK_FAILURE_MESSAGE = "Cloudflare tunnel failed to connect. This is likely due to firewall/proxy blocking the connection."
    _STARTUP_ERROR_MESSAGE = "Cloudflare tunnel encountered errors during startup."

    @staticmethod
    def _validate_cloudflare_config(config_data: Dict[str, Any], service_config: ServiceConfig) -> None:
        """Validate Cloudflare configuration parameters."""
        required_keys = ["port", "cloudflare_tunnel_token", "cloudflare_custom_domain", "run_mode"]
        
        for key in required_keys:
            if key not in config_data:
                raise ValidationError(f"Missing required configuration key: {key}")
        
        service_config.validator.validate_port(config_data["port"])
        service_config.validator.validate_cloudflare_token(config_data["cloudflare_tunnel_token"])
        service_config.validator.validate_domain(config_data["cloudflare_custom_domain"])
        
        if config_data["run_mode"] not in ["foreground", "background"]:
            raise ValidationError(f"Invalid run_mode: {config_data['run_mode']}")
        
        logger.debug("Cloudflare configuration validation successful")
    
    @staticmethod
    def _check_cloudflare_tunnel_health(max_wait_seconds=None):
        """
        Check if cloudflare tunnel started successfully by examining the log file.
        Returns (success: bool, error_message: str)
        """
        max_wait_seconds = max_wait_seconds or CloudflareConfigurator._DEFAULT_HEALTH_CHECK_TIMEOUT
        log_file = CloudflareConfigurator._get_cloudflare_log_path()
        
        for _ in range(max_wait_seconds):
            if os.path.exists(log_file):
                success, error = CloudflareConfigurator._analyze_tunnel_log(log_file)
                if success is not None:  # Definitive result
                    return success, error
            time.sleep(CloudflareConfigurator._HEALTH_CHECK_RETRY_INTERVAL)
        
        return False, CloudflareConfigurator._SUCCESS_TIMEOUT_MESSAGE

    @staticmethod
    def _get_cloudflare_log_path() -> str:
        """Get the path to the Cloudflare tunnel log file."""
        service_core_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "core")
        return os.path.join(service_core_dir, "logs", "cloudflare_tunnel_subprocess.log")

    @staticmethod
    def _analyze_tunnel_log(log_file: str) -> tuple[Optional[bool], str]:
        """
        Analyze tunnel log content for success/failure indicators.
        Returns (success: Optional[bool], error_message: str)
        - True: Success detected
        - False: Failure detected  
        - None: No definitive result yet
        """
        try:
            content = CloudflareConfigurator._read_log_file(log_file)
            return CloudflareConfigurator._check_tunnel_patterns(content)
        except (IOError, OSError, UnicodeDecodeError) as e:
            logger.debug(f"Error reading cloudflare log file: {type(e).__name__}")
            return None, ""
        except Exception as e:
            logger.error(f"Unexpected error reading cloudflare log: {type(e).__name__}")
            raise

    @staticmethod
    def _read_log_file(log_file: str) -> str:
        """Read and return the content of the log file."""
        with open(log_file, 'r', encoding='utf-8') as f:
            return f.read()

    @staticmethod
    def _check_tunnel_patterns(content: str) -> tuple[Optional[bool], str]:
        """Check log content for success/failure patterns."""
        if any(pattern in content for pattern in CloudflareConfigurator._SUCCESS_PATTERNS):
            return True, ""
            
        if any(pattern in content for pattern in CloudflareConfigurator._FAILURE_PATTERNS):
            return False, CloudflareConfigurator._NETWORK_FAILURE_MESSAGE
            
        if "ERR" in content and "Failed" in content:
            return False, CloudflareConfigurator._STARTUP_ERROR_MESSAGE
            
        return None, ""

    @staticmethod
    @debug_decorator  
    def configure_cloudflare(config_data: Dict[str, Any], service_config: ServiceConfig) -> Optional[int]:
        """Configure Cloudflare tunnel if enabled. Returns tunnel PID if started in background mode, None otherwise."""
        if config_data.get("cloudflare") != 'y':
            return None
        
        logger.debug("Configuring Cloudflare tunnel")
        
        try:
            CloudflareConfigurator._validate_cloudflare_config(config_data, service_config)
            
            result = generate_cloudflare_url(
                config_data["port"], 
                config_data["cloudflare_tunnel_token"],
                config_data["cloudflare_custom_domain"],
                config_data["run_mode"],
            )
            
            if isinstance(result, tuple):
                config_data["cloudflare_public_url"], cloudflare_pid = result
                
                # Check if tunnel started successfully
                logger.debug("Checking Cloudflare tunnel health...")
                tunnel_success, error_message = CloudflareConfigurator._check_cloudflare_tunnel_health()
                
                if not tunnel_success:
                    CloudflareConfigurator._cleanup_failed_process(cloudflare_pid)
                    raise Exception(f"Commander Service failed to start: {error_message}")
                
                if config_data["cloudflare_public_url"]:
                    print(f'Generated Cloudflare tunnel URL: {config_data["cloudflare_public_url"]}')
                else:
                    print('Cloudflare tunnel started, URL will be available in logs')
                return cloudflare_pid
            else:
                config_data["cloudflare_public_url"] = result
                print(f'Generated Cloudflare tunnel URL: {result}')
                return None
                
        except ValidationError as e:
            logger.error(f"Invalid Cloudflare configuration: {e}")
            raise
        except Exception as e:
            logger.error(f"Failed to configure Cloudflare tunnel: {e}")
            raise

    @staticmethod
    def _cleanup_failed_process(cloudflare_pid: Optional[int]) -> None:
        """Clean up failed Cloudflare process."""
        if not cloudflare_pid:
            return
            
        logger.debug(f"Cleaning up failed Cloudflare process {cloudflare_pid}")
        
        try:
            if psutil:
                try:
                    process = psutil.Process(cloudflare_pid)
                    process.terminate()
                    logger.debug(f"Terminated failed cloudflare process {cloudflare_pid}")
                except (psutil.NoSuchProcess, psutil.AccessDenied) as e:
                    logger.debug(f"Process cleanup - process not found or access denied: {type(e).__name__}")
                except psutil.ZombieProcess:
                    logger.debug(f"Process {cloudflare_pid} is already a zombie process")
                except (OSError, IOError) as e:
                    logger.warning(f"OS error during process cleanup: {type(e).__name__}")
            else:
                logger.warning("Cannot terminate cloudflare process: psutil not available")
        except (KeyboardInterrupt, SystemExit):
            logger.info("Process cleanup interrupted by user or system")
            raise  
        except Exception as e:
            logger.error(f"Unexpected error during process cleanup: {type(e).__name__}")

